# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Process lobulus analysis.
"""
# import logging
# logger = logging.getLogger(__name__)
from loguru import logger
import sys
import subprocess
import os


def open_path(path):
    """
    Open path (file or dir) in default operation system application.

    :param path:
    :return:
    """
    if sys.platform.startswith("darwin"):
        cmd = "open"
        # subprocess.call(('open', path))
    elif os.name == "nt":  # For Windows
        cmd = "start"
        # os.startfile(filepath)
    elif os.name == "posix":  # For Linux, Mac, etc.
        cmd = "xdg-open"
        # subprocess.call(('xdg-open', filepath))
    else:
        raise EnvironmentError("Unknown OS: " + os.name)

    try:
        retcode = subprocess.call(cmd + " " + str(path), shell=True)
        if retcode < 0:
            logger.error(f"Child was terminated by signal {retcode}",)
        else:
            logger.info(f"Child returned {retcode}")
    except OSError as e:
        logger.error("Execution failed: %s", e)
        raise e



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Process lobulus analysis.
"""

from loguru import logger
import skimage.filters
from skimage.morphology import skeletonize
import skimage.io
import scipy.signal
import scipy.ndimage
import os.path as op
import numpy as np
import warnings
import morphsnakes as ms
from matplotlib import pyplot as plt
from scaffan import image as scim
import scaffan
import scaffan.lobulus
from exsu.report import Report
from pyqtgraph.parametertree import Parameter
import imma.image


class SkeletonAnalysis:
    def __init__(
        self,
        pname="Skeleton Analysis",
        ptype="bool",
        pvalue=True,
        ptip="Skeleton Analysis after lobulus segmentation is performed",
    ):

        params = [
            # {
            #     "name": "Tile Size",
            #     "type": "int",
            #     "value" : 128
            # },
            {
                "name": "Working Resolution",
                "type": "float",
                # "value": 0.000001,
                "value": 0.00000091,  # this is typical resolution on level 2
                # "value": 0.00000182,  # this is typical resolution on level 3
                # "value": 0.00000364,  # this is typical resolution on level 4
                # "value": 0.00000728,  # this is typical resolution on level 5
                # "value": 0.00001456,  # this is typical resolution on level 6
                "suffix": "m",
                "siPrefix": True,
            },
            {
                "name": "Inner Lobulus Margin",
                "type": "float",
                "value": 0.00002,
                "suffix": "m",
                "siPrefix": True,
                "tip": "Area close to the border is ignored in Otsu threshold computation before skeletonization step.",
            },
        ]

        self.parameters = Parameter.create(
            name=pname,
            type=ptype,
            value=pvalue,
            tip=ptip,
            children=params,
            expanded=False,
        )
        self.report: Report = None

    def set_report(self, report: Report):
        self.report = report

    def set_lobulus(self, lobulus: scaffan.lobulus.Lobulus):
        self.lobulus: scaffan.lobulus.Lobulus = lobulus
        pixelsize_mm = [
            (self.parameters.param("Working Resolution").value() * 1000)
        ] * 2
        self.view = self.lobulus.view.to_pixelsize(pixelsize_mm=pixelsize_mm)
        logger.debug("Lobulus setup done")

    def skeleton_analysis(self, show=False):
        datarow = {}

        inner = self.lobulus.central_vein_mask
        logger.debug(
            f"dtype of lobulus.central_vein_mask {self.lobulus.central_vein_mask.dtype}"
        )
        logger.debug(
            f"shape of lobulus.central_vein_mask {self.lobulus.central_vein_mask.shape} {np.max(self.lobulus.central_vein_mask)}"
        )
        logger.debug(
            f"unique of shape of lobulus.central_vein_mask {np.unique(self.lobulus.central_vein_mask)}"
        )
        # import pdb; pdb.set_trace()
        # TODO Split the function here
        inner_lobulus_margin_mm = (
            self.parameters.param("Inner Lobulus Margin").value() * 1000
        )

        # eroded image for threshold analysis
        dstmask = scipy.ndimage.morphology.distance_transform_edt(
            self.lobulus.lobulus_mask, self.lobulus.view.region_pixelsize[::-1]
        )
        inner_lobulus_mask = dstmask > inner_lobulus_margin_mm
        # print("inner_lobulus_mask" , np.sum(inner_lobulus_mask==0), np.sum(inner_lobulus_mask==1))

        # detail_level = 2
        new_size = self.view.get_size_on_pixelsize_mm()

        resize_params = dict(
            output_shape=[new_size[1], new_size[0]],
            mode="reflect",
            order=0,
            anti_aliasing=False,
        )
        detail_mask = skimage.transform.resize(
            self.lobulus.lobulus_mask, **resize_params
        ).astype(np.int8)
        detail_inner_lobulus_mask = skimage.transform.resize(
            inner_lobulus_mask, **resize_params
        )
        detail_central_vein_mask = skimage.transform.resize(
            inner == 1, **resize_params
        ).astype(np.int8)

        detail_view = self.view
        detail_image = detail_view.get_region_image(as_gray=True)
        fig = plt.figure()
        plt.imshow(detail_image)
        plt.contour(detail_mask + detail_inner_lobulus_mask)
        detail_view.add_ticks()
        if self.report is not None:
            self.report.savefig_and_show(
                "skeleton_analysis_detail_image_and_mask_{}.png".format(
                    self.lobulus.annotation_id
                ),
                fig,
            )

        # from PyQt5.QtCore import pyqtRemoveInputHook
        # pyqtRemoveInputHook()

        # Or for Qt5
        # from PyQt5.QtCore import pyqtRemoveInputHook

        # from pdb import set_trace
        # set_trace()
        logger.debug(
            f"detail_inner_lobulus_mask {detail_central_vein_mask.shape} {detail_central_vein_mask.dtype} "
            f"{np.min(detail_central_vein_mask)} {np.max(detail_central_vein_mask)}"
        )
        threshold = skimage.filters.threshold_otsu(
            detail_image[detail_inner_lobulus_mask == 1]
        )
        imthr = detail_image < threshold
        imthr[detail_mask != 1] = 0
        # plt.figure()
        # plt.imshow(imthr)
        # if show:
        #     plt.show()
        skeleton = skeletonize(imthr)
        datarow["Skeleton length"] = np.sum(skeleton) * detail_view.region_pixelsize[0]
        datarow["Output pixel size 0"] = detail_view.region_pixelsize[0]
        datarow["Output pixel size 1"] = detail_view.region_pixelsize[1]
        datarow["Output image size 0"] = (
            detail_view.region_pixelsize[0] * imthr.shape[0]
        )
        datarow["Output image size 1"] = (
            detail_view.region_pixelsize[1] * imthr.shape[1]
        )
        fig = plt.figure(figsize=(12, 10))
        plt.imshow(skeleton + imthr)
        detail_view.add_ticks()
        if self.report is not None:
            self.report.savefig_and_show(
                "thumb_skeleton_thr_{}.png".format(self.lobulus.annotation_id), fig
            )

        if self.report is not None:
            imall = detail_mask.astype(np.uint8)
            imall[detail_central_vein_mask > 0] = 2
            imall[imthr > 0] = 3
            imall[skeleton > 0] = 4
            # imall = (skeleton.astype(np.uint8) + imthr.astype(np.uint8) +  + (detail_central_vein_mask.astype(np.uint8) * 2)).astype(np.uint8)
            self.imsave("lobulus_central_thr_skeleton_{}.png", imall)
            self.imsave(
                "lobulus_thr_skeleton_{}.png",
                (skeleton.astype(np.uint8) + imthr + detail_mask).astype(np.uint8),
            )
            self.imsave("skeleton_{}.png", skeleton)
            self.imsave("thr_{}.png", imthr)
            # plt.imsave(op.join(self.report.outputdir, "skeleton_thr_lobulus_{}.png".format(self.annotation_id)), skeleton.astype(np.uint8) + imthr + detail_mask)
            # plt.imsave(op.join(self.report.outputdir, "skeleton_{}.png".format(self.annotation_id)), skeleton)
            # plt.imsave(op.join(self.report.outputdir, "thr_{}.png".format(self.annotation_id)), imthr)
            # skimage.io.imsave(op.join(self.report.outputdir, "raw_skeleton_thr_lobulus_{}.png".format(self.annotation_id)),
            #                   (50 * skeleton + 50 * imthr + 50 * detail_mask).astype(np.uint8))
            # skimage.io.imsave(op.join(self.report.outputdir, "raw_skeleton_{}.png".format(self.annotation_id)), 50 * skeleton)
            # skimage.io.imsave(op.join(self.report.outputdir, "raw_thr_{}.png".format(self.annotation_id)), 50 * imthr)

        conv = scipy.signal.convolve2d(skeleton, np.ones([3, 3]), mode="same")
        conv = conv * skeleton
        fig = plt.figure(figsize=(12, 10))
        plt.imshow(conv)
        detail_view.add_ticks()
        if self.report is not None:
            self.report.savefig_and_show(
                "figure_skeleton_nodes_{}.png".format(self.lobulus.annotation_id), fig
            )

            with warnings.catch_warnings():
                # warnings.simplefilter("low contrast image")
                warnings.filterwarnings("ignore", ".*low contrast image.*")
                self.imsave("skeleton_nodes_raw_{}.png", conv, 20)
            # plt.imsave(op.join(self.report.outputdir, "skeleton_nodes_{}.png".format(self.annotation_id)), conv.astype(np.uint8))
            # skimage.io.imsave(op.join(self.report.outputdir, "raw_skeleton_nodes_{}.png".format(self.annotation_id)), (conv * 20).astype(np.uint8))

        conv[conv > 3] = 0
        label, num = scipy.ndimage.label(conv)
        datarow["Branch number"] = num
        label, num = scipy.ndimage.label(conv == 1)
        datarow["Dead ends number"] = num
        if "Area" in datarow:
            area_unit = datarow["Area unit"]
            datarow[f"Branch number density [1/{area_unit}^2]"] = datarow["Branch number"]/datarow["Area"]
            datarow[f"Dead ends number density [1/{area_unit}^2]"] = datarow["Dead ends number"]/datarow["Area"]
            datarow[f"Skeleton length density [{area_unit}/{area_unit}^2]"] = datarow["Branch length number"]/datarow["Area"]
        else:
            # probably area can be estimated by view area
            logger.debug("Unknown area. Skipping density calculation")

        self.report.add_cols_to_actual_row(datarow)

    def imsave(self, base_fn, arr, k=50):
        base_fn = base_fn.format(self.lobulus.annotation_id)
        self.report.imsave(base_fn, arr, k)



import pyqtgraph


def params_and_values(
    p: pyqtgraph.parametertree.Parameter, pth=None, dct={}, separator=";"
):
    """
    Get dict of all parameters. Key is the path to the parameter, value is value of the parameter.
    :param p:
    :param pth:
    :param dct:
    :param separator:  default ";"
    :return:
    """
    for name in p.getValues():
        # print(f"name: {name}, type {type(name)}")
        if pth is not None:
            pth_local = pth + separator + name
        else:
            pth_local = name
        # print(pth)
        ch = p.child(name)
        # print(f"name: {name}, type {type(ch)}")
        if type(ch) is pyqtgraph.parametertree.parameterTypes.SimpleParameter:
            dct[pth_local] = ch.value()
            # print(pth)
        else:
            params_and_values(ch, pth_local)

    return dct



# /usr/bin/env python
# -*- coding: utf-8 -*-
from loguru import logger

# import scaffan
# import io3d # just to get data
import scaffan.image as scim
from typing import List, Union
from pathlib import Path

# import sklearn.cluster
# import sklearn.naive_bayes
# import sklearn.svm
from scaffan.image import View
from scaffan.annotation import annotation_px_to_mm

# from sklearn.externals import joblib
import joblib
from scipy.ndimage import gaussian_filter
import skimage
from sklearn.naive_bayes import GaussianNB
import numpy as np
from skimage.feature import peak_local_max
import skimage.filters
import sklearn.metrics
# import skimage.metrics
from skimage.morphology import disk
import scipy.ndimage
import matplotlib.pyplot as plt
from pyqtgraph.parametertree import Parameter, ParameterTree
from sklearn.neural_network import MLPClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
import exsu
from exsu.report import Report
from .image import AnnotatedImage
from . import texture
from . import whole_slide_seg_unet
import scaffan


class ScanSegmentation:
    def __init__(
        self,
        report: Report = None,
        pname="Scan Segmentation",
        ptype="bool",
        pvalue=True,
        ptip="Run analysis of whole scan before all other processing is perfomed.\n"
        + "If automatic lobulus selection is selected, "
        + "defined number of biggest lobuli are selected for texture analysis.",
    ):
        """

        An alternative features computation can be done by setting callback function
        self.alternative_get_features = fcn(seg:SlideSegmentation, view:View) -> ndarray:

        :param report:
        :param ptype: group or bool
        """
        # self._inner_texture:texture.GLCMTextureMeasurement = texture.GLCMTextureMeasurement(
        #     "slide_segmentation", texture_label="slide_segmentation", report_severity_offset=-30,
        #     glcm_levels=128
        # )
        self._unet = whole_slide_seg_unet.WholeSlideSegmentationUNet(report=report)

        self._glcm = scaffan.texture.GLCMTextureMeasurement(
            "whole_slide_glcm",
            texture_label="ws_part",
            add_cols_to_report=False,
            report=report,
            # tile_size=64,
            # tile_spacing=32,
            tile_size=16,
            tile_spacing=8,
            report_severity_offset=-30
        )
        params = [
            # {
            #     "name": "Tile Size",
            #     "type": "int",
            #     "value" : 128
            # },
            {
                "name": "Working Resolution",
                "type": "float",
                # "value": 0.00001,  # 0.01 mm
                "value": 0.000005,  # 0.005 mm
                "value": 0.00000355,  # 0.005 mm
                "suffix": "m",
                "siPrefix": True,
                "tip": "Resolution used for segmentation processing. "
                + "Real working resolution will be selected according to image levels.",
            },
            {
                "name": "Working Tile Size",
                "type": "int",
                "value": 256,
                "suffix": "px",
                "siPrefix": False,
                "tip": "Image is processed tile by tile. This value defines size of the tile",
            },
            {
                "name": "Segmentation Method",
                "type": "list",
                "values": ["U-Net", "HCTFS", "GLCMTFS"],
                "value": "GLCMTFS",
                # "value": "U-Net",
            },
            self._unet.parameters,
            {
                "name": "TFS General",
                "type": "group",
                "tip": "Texture Features based Segmentation parameters",
                "expanded": False,
                "children": [
                    {
                        "name": "Run Training",
                        "type": "bool",
                        "value": False,
                        "tip": "Use annotated image to train classifier.\n"
                        + "Red (#FF0000) area is extra-lobular tissue.\n"
                        + "Black (#000000) area is intra-lobular tissue.\n"
                        + "Magenta (#FF00FF) area is empty part of the image.\n",
                    },
                    {
                        "name": "Load Default Classifier",
                        "type": "bool",
                        "value": False,
                        "tip": "Load default classifier before training and prediction.",
                    },
                    {
                        "name": "Clean Before Training",
                        "type": "bool",
                        "value": False,
                        "tip": "Reset classifier before training.",
                    },
                    {
                        "name": "Training Weight",
                        "type": "float",
                        "value": 1,
                        # "suffix": "px",
                        "siPrefix": False,
                        "tip": "Weight of training samples given in actual image",
                    },
                    {
                        "name": "Training Stride",
                        "type": "int",
                        "value": 1,
                        "suffix": "px",
                        "siPrefix": False,
                        "tip": "Every n-th pixel is used for training. Push down the computation expences. Usefull on debug.",
                    },
                ],
            },
            {
                "name": "Save Training Labels",
                "type": "bool",
                "value": False,
                # "suffix": "px",
                "siPrefix": False,
                "tip": "Training labels from original image are saved to output dir. "
                + "There is computation cost when it is turned on. "
                + "It is not used in standard processing. "
                + "The training labels are: 'unknown', 'background', 'intralobular' and 'interlobular'.",
            },
            {
                "name": "Postprocessing",
                "type": "bool",
                "value": True,
                "tip": "Run postprocessing based on intra-lobular smoothing and maximal distance from extra-lobular tissue",
                "expanded": False,
                "children": [
                    {
                        "name": "Max. Distance",
                        "type": "float",
                        "value": 0.001,  # 1 mm
                        # "value": 0.000005,  # 0.005 mm
                        # "value": 0.00000355,  # 0.005 mm
                        "suffix": "m",
                        "siPrefix": True,
                        "tip": "Maximal distance from extra-lobular tissue"
                    },
                ]
            },
            {
                "name": "Lobulus Number",
                "type": "int",
                "value": 5,
                # "suffix": "px",
                "siPrefix": False,
                "tip": "Number of lobuluses automatically selected after whole scan segmentation",
            },
            {
                "name": "Annotation Radius",
                "type": "float",
                "value": 0.00015,  # 0.1 mm
                "suffix": "m",
                "siPrefix": True,
                "tip": "Radius of circle seed used as input for individual lobulus segmentation "
                + "when the automatic lobulus selection is prefered ",
            },
            {
                "name": "Run Prediction",
                "type": "bool",
                "value": True,
                # "suffix": "px",
                "siPrefix": False,
                "tip": "Per-pixel prediction (segmentation) is the key step in slide segmentation.\n" +
                "It can be skipped during the training phase or for debug reasons."
            },
            # self._inner_texture.parameters,
        ]

        self.parameters = Parameter.create(
            name=pname,
            type=ptype,
            value=pvalue,
            tip=ptip,
            children=params,
            expanded=False,
        )
        if report is None:
            report = Report()
            report.save = False
            report.show = False
        self.report: Report = report
        self.anim: AnnotatedImage = None
        self.tile_size = None
        self.level = None
        self.tiles: List["View"] = None
        # self._clf_object = SVC
        # self._clf_params = dict(gamma=2, C=1)
        self._clf_object = GaussianNB
        self._clf_params = {}
        self.whole_slide_training_labels = None
        self.compatible_with_openslide = True
        # self._clf_object = DecisionTreeClassifier # no partial fit :-(
        # self._clf_params = dict(max_depth=5)

        #         self.clf = sklearn.svm.SVC(gamma='scale')
        # KNeighborsClassifier(3),
        # SVC(kernel="linear", C=0.025),
        # SVC(gamma=2, C=1),
        # #     GaussianProcessClassifier(1.0 * RBF(1.0)),
        # self.clf_fn = DecisionTreeClassifier(max_depth=5),
        # RandomForestClassifier(max_depth=5, n_estimators=10, max_features=1),
        # MLPClassifier(alpha=1, max_iter=1000),
        # AdaBoostClassifier(),
        # QuadraticDiscriminantAnalysis(),
        # GaussianNB(),
        # self.clf = GaussianNB()
        self.clf = None
        self.clf_fn = None
        self.clf_default_fn = None
        self.predicted_tiles = None
        # self.output_label_fn = "label.png"
        # self.output_raster_fn = "image.png"
        self.devel_imcrop = None
        #         self.devel_imcrop = np.array([20000, 15000])
        self.full_output_image = None
        self.full_raster_image = None
        self.used_pixelsize_mm = None
        self.ann_biggest_ids = []
        self.empty_area_mm = None
        self.septum_area_mm = None
        self.sinusoidal_area_mm = None
        self.alternative_get_features = None
        self.accuracy = None

    # def init(self, anim: scim.AnnotatedImage):
    def init(self, view: scim.View):
        """
        :param view: full view can be extracted by anim.get_full_view
        :return:
        """
        self.set_view(view)
        # self.anim = anim # is set in set_view()
        # self.anim = scim.AnnotatedImage(str(fn))
        self.level = self._find_best_level()
        self.tiles = None
        self.predicted_tiles = None
        self.tile_size = None
        self.ann_biggest_ids = []
        self.make_tiles()
        self.init_clf()

    def init_clf(self):
        method = str(self.parameters.param("Segmentation Method").value())
        logger.debug(f"method={method}")
        method_str = method.replace(" ", "_")
        self.clf = self._clf_object(**self._clf_params)
        self.clf_fn: Path = Path(Path(__file__).parent / f"segmentation_model_{method_str}.pkl")
        self.clf_default_fn: Path = Path(
            Path(__file__).parent / f"segmentation_model_default_{method_str}.pkl"
        )
        if self.clf_fn.exists():
            logger.debug(f"Reading classifier from {str(self.clf_fn)}")
            self.clf = joblib.load(self.clf_fn)

    def train_classifier(self, pixels=None, y=None, sample_weight: float = None):
        logger.debug("start training")

        if pixels is None:
            pixels, y = self.prepare_training_pixels()
        if sample_weight is None:
            sample_weight = float(
                self.parameters.param("TFS General", "Training Weight").value()
            )
        sample_weight = [sample_weight] * len(y)

        if bool(self.parameters.param("TFS General", "Clean Before Training").value()):
            self.clf = self._clf_object(**self._clf_params)
            # self.clf = GaussianNB()
            logger.debug(f"cleaning the classifier {self.clf}")
            self.clf.fit(pixels, y=y, sample_weight=sample_weight)
        else:
            self.clf.partial_fit(pixels, y=y, sample_weight=sample_weight)
        logger.debug("training finished")

    def save_classifier(self):
        logger.debug(f"save clf to {self.clf_fn}")
        joblib.dump(self.clf, self.clf_fn)

    def prepare_training_pixels(self):
        """
        Use annotated image to train classifier.
        Red area is extra-lobular tissue.
        Black area is intra-lobular tissue.
        Magenta area is empty part of the image.
        """
        pixels0 = self._get_pixels("#FF00FF")  # empty
        pixels1 = self._get_pixels("#000000")  # black
        pixels2 = self._get_pixels("#FF0000")  # extra lobula
        labels0 = np.ones([pixels0.shape[0]]) * 0
        labels1 = np.ones([pixels1.shape[0]]) * 1
        labels2 = np.ones([pixels2.shape[0]]) * 2
        pixels = np.concatenate([pixels0, pixels1, pixels2])
        y = np.concatenate([labels0, labels1, labels2])

        return pixels, y

    def _get_pixels(self, color: str):
        """
        Use outer annotation with defined color and removed holes to
        extract features in pixels.
        """
        outer_ids, holes_ids = self.anim.select_just_outer_annotations(color)
        views = self.anim.get_views(outer_ids, level=self.level)
        pixels_list = []
        for id1, id2, view_ann in zip(outer_ids, holes_ids, views):
            ann_raster = view_ann.get_annotation_raster(id1, holes_ids=id2)
            #     ann_raster1 = view_ann.get_annotation_region_raster(id1)
            #     if len(id2) == 0:
            #         ann_raster = ann_raster1
            #     else:
            #         ann_raster2 = view_ann.get_annotation_region_raster(id2[0])
            #         ann_raster = ann_raster1 ^ ann_raster2

            #             plt.figure()
            #             plt.imshow(ann_raster)
            #             plt.show()
            img = self._get_features(view_ann, annotation_id=id1)
            stride = int(self.parameters.param("TFS General", "Training Stride").value())
            pixels = img[ann_raster][::stride]
            pixels_list.append(pixels)
        pixels_all = np.concatenate(pixels_list, axis=0)
        return pixels_all

    def _get_features(self, view: View, debug_return=False, annotation_id=None) -> np.ndarray:
        method = str(self.parameters.param("Segmentation Method").value())
        if method == "HCTFS":
            return self._get_features_hctf(view, debug_return=debug_return)
        elif method == "GLCMTFS":
            texture_label_str_id = f"{view.region_location[0]}_{view.region_location[1]}" if annotation_id is None else None
            self._glcm.set_input_data(
                view=view, annotation_id=annotation_id, lobulus_segmentation=None,
                texture_label_str_id=texture_label_str_id

            )
            self._glcm.run(recalculate_view=False)
            features = self._glcm.measured_features
            if debug_return:
                return features, []
            else:
                return features
        pass

    def _get_features_hctf(self, view: View, debug_return=False) -> np.ndarray:
        """
        Three colors and one gaussian smooth reg channel.
        An alternative features computation can be done by setting callback function
        self.alternative_get_features = fcn(seg:SlideSegmentation, view:View) -> ndarray:

        img_sob: gaussian blure applied on gradient sobel operator give information about texture richness in neighborhood

        """
        img = view.get_region_image(as_gray=False)
        img_gauss2 = gaussian_filter(img[:, :, 0], 2)
        img_gauss5 = gaussian_filter(img[:, :, 0], 5)

        img = np.copy(img)
        nfeatures = 9
        # nfeatures = 12 #GLCM
        imgout = np.zeros([img.shape[0], img.shape[1], nfeatures], dtype=np.uint8)
        img_just_sob = skimage.filters.sobel(img[:, :, 0])
        # print(f"just_sob {img_just_sob.dtype} stats: {stats.describe(img_just_sob[:], axis=None)}")
        #         img_sob = (np.abs(img_just_sob) * 255).astype(np.uint8)
        img_sob = (np.abs(img_just_sob) * 255).astype(np.uint8)
        # print(f"sob {img_sob.dtype} stats: {stats.describe(img_sob[:], axis=None)}")
        img_sob_gauss2 = gaussian_filter(img_sob, 2)
        img_sob_gauss5 = gaussian_filter(img_sob, 5)
        img_sob_median = skimage.filters.median(
            (img_just_sob * 2000).astype(np.uint8), disk(10)
        )

        # GLCM
        # self._inner_texture.set_input_data(view=view, annotation_id=None, lobulus_segmentation=None)
        # self._inner_texture.run(recalculate_view=False)
        # glcm_features = (self._inner_texture.measured_features * 255).astype(np.uint8)
        # imgout[:, :, 9:12] = glcm_features[:, :, :3]

        imgout[:, :, :3] = img[:, :, :3]
        imgout[:, :, 3] = img_gauss2
        imgout[:, :, 4] = img_gauss5
        imgout[:, :, 5] = img_sob
        imgout[:, :, 6] = img_sob_gauss2
        imgout[:, :, 7] = img_sob_gauss5
        imgout[:, :, 8] = img_sob_median

        if debug_return:
            return imgout, [img_sob]
        return imgout

    def _find_best_level(self):
        pixelsize_mm = np.array(
            [float(self.parameters.param("Working Resolution").value()) * 1000] * 2
        )
        logger.debug(f"wanted pixelsize mm={pixelsize_mm}")
        error = None
        closest_i = None
        best_pxsz = None
        for i, pxsz in enumerate(self.anim.level_pixelsize):
            err = np.linalg.norm(pixelsize_mm - pxsz)
            if error is None:
                error = err
                closest_i = i
                best_pxsz = pxsz
            else:
                if err < error:
                    error = err
                    closest_i = i
                    best_pxsz = pxsz
        self.used_pixelsize_mm = best_pxsz
        logger.debug(f"real pixelsize mm={best_pxsz}")

        return closest_i

    def set_view(self, view: View):
        self.view = view.to_level(0)
        self.anim = view.anim

    def _get_tiles_parameters(self):
        sp_height0, sp_width0 = self.view.get_size_on_level(0)
        st_height0, st_width0 = self.view.region_location  # pozor, tady je to prohozený

        imcrop_height0 = sp_height0 + 1 * st_height0
        imcrop_width0 = sp_width0 + 1 * st_width0
        height0, width0 = self.view.get_size_on_level(0)

        # height_check = self.anim.openslide.properties["openslide.level[0].height"]
        # width_check = self.anim.openslide.properties["openslide.level[0].width"]

        # imsize = np.array([int(width0), int(height0)])
        imsize_on_level0 = np.array([int(height0), int(width0)])
        cr_sp_on_level0 = np.array([int(imcrop_height0), int(imcrop_width0)])
        if self.devel_imcrop is not None:
            imsize_on_level0 = self.devel_imcrop

        tile_size_on_level = np.array(self.tile_size)
        downsamples = self.anim.openslide.level_downsamples[self.level]
        imsize_on_level = imsize_on_level0 / downsamples
        tile_size_on_level0 = tile_size_on_level * downsamples
        self._imsize_on_level = imsize_on_level
        self._imsize_on_level0 = imsize_on_level0.astype(np.int)
        self._tile_size_on_level0 = tile_size_on_level0.astype(np.int)
        self._tile_size_on_level = tile_size_on_level.astype(np.int)
        self._sl_sz_on_level = imsize_on_level.astype(np.int)
        self._cr_sp_on_level0 = cr_sp_on_level0.astype(np.int)
        self._cr_sp_on_level = (cr_sp_on_level0 / downsamples).astype(np.int)
        self._cr_st_on_level0 = np.asarray([st_height0, st_width0], dtype=np.int)
        self._cr_st_on_level = (self._cr_st_on_level0 / downsamples).astype(np.int)
        self._cr_sz_on_level = self._cr_sp_on_level - self._cr_st_on_level
        self._cr_sz_on_level0 = self._cr_sp_on_level0 - self._cr_st_on_level0

        return (
            imsize_on_level0.astype(np.int),
            tile_size_on_level0.astype(np.int),
            tile_size_on_level,
            imsize_on_level.astype(np.int),
        )

    def make_tiles(self):
        sz = int(self.parameters.param("Working Tile Size").value())
        self.tile_size = [sz, sz]
        (
            imsz_on_level0,
            size_on_level0,
            size_on_level,
            imsz_on_level,
        ) = self._get_tiles_parameters()
        logger.debug(
            f"level={self.level}, tile size on level={size_on_level}, tile size on level 0={size_on_level0}"
        )
        # self.tiles = []

        # for x0 in range(0, int(imsz_on_level0[0]), int(size_on_level0[0])):
        #     column_tiles = []
        #
        #     for y0 in range(0, int(imsz_on_level0[1]), int(size_on_level0[1])):
        #         logger.trace(f"processing tile {x0}, {y0}")
        #         view = self.anim.get_view(
        #             location=(x0, y0), size_on_level=size_on_level, level=self.level
        #         )
        #         column_tiles.append(view)
        #
        #     self.tiles.append(column_tiles)

        # todo the iterator is strange the x and y seems to be swapped sometimes
        self.tiles = []
        column_tiles = []
        for tile_params in self.tile_iterator(
            return_in_out_coords=False,
            return_tile_coords=True,
            return_level0_coords=True,
        ):
            sl_tl, sl_gl, (x0, y0) = tile_params
            # ix, iy, x0, y0 = tile_params
            # if len(self.tiles2) <= ix:
            #     column_tiles = []
            #     self.tiles2.append(column_tiles)
            view = self.anim.get_view(
                location=(x0, y0), size_on_level=size_on_level, level=self.level
            )
            self.tiles.append((view, tile_params))
        pass

    def predict_on_view(self, view):
        smethod = str(self.parameters.param("Segmentation Method").value())
        if smethod in ["HCTFS", "GLCMTFS"]:
            return self.predict_on_view_tfs(view)
        elif smethod == "U-Net":
            return self._unet.predict_tile(view)
        else:
            raise ValueError(f"Unknown segmentation method '{smethod}'")

    def predict_on_view_tfs(self, view):
        image = self._get_features(view)
        fvs = image.reshape(-1, image.shape[2])
        #         print(f"fvs: {fvs[:10]}")
        predicted = self.clf.predict(fvs).astype(np.int)
        img_pred = predicted.reshape(image.shape[0], image.shape[1])
        return img_pred

    def save_training_labels(self):
        if self.tiles is None:
            self.make_tiles()
        logger.debug("saving training labels")
        # (
        #     imsize,
        #     tile_size_on_level0,
        #     tile_size_on_level,
        #     imsize_on_level,
        # ) = self._get_tiles_parameters()
        # (
        imsize = self._imsize_on_level0
        tile_size_on_level0 = self._tile_size_on_level0
        tile_size_on_level = self._tile_size_on_level
        imsize_on_level = self._sl_sz_on_level
        # ) = self._get_tiles_parameters()
        # szx = len(self.tiles)
        # szy = len(self.tiles[0])
        output_image = np.zeros(self.tile_size + imsize_on_level, dtype=int)

        # for sl_x_tl, sl_y_tl, sl_x_gl, sl_y_gl, ix, iy  in self.tile_iterator(return_tile_coords=True):
        for view, tile_params in self.tiles:
            # view = self.tiles[ix][iy]
            sl_tl, sl_gl, loc = tile_params
            segmentation = view.get_training_labels()
            output_image[
                sl_gl
            ] = segmentation  # .get_region_image(as_gray=as_gray)[sl_x_tl, sl_y_tl, :3]

        self.whole_slide_training_labels = output_image[self._full_image_crop_slice()]
        self.report.imsave(
            "whole_slide_training_labels.png",
            self.whole_slide_training_labels,
            level_skimage=40,
            level_npz=30,
        )

    def _imsave_full_raster_and_training_labels(self):
        if (self.full_raster_image is not None) and (self.whole_slide_training_labels is not None):
            fig = plt.figure()
            plt.imshow(self.full_raster_image)
            plt.contour(self.whole_slide_training_labels,
                        # cmap="Purples"
                        )
            self.report.savefig("whole_slide_raster_and_training_labels.png", level=45)
            plt.close(fig)

    def predict_tiles(self):
        if self.tiles is None:
            self.make_tiles()

        logger.debug("predicting tiles")
        self.predicted_tiles = []
        for i, (tile_view, tile_params) in enumerate(self.tiles):
            # for i, tile_view_col in enumerate(self.tiles):
            # logger.trace(f"predicting tiles in {i}-th row")
            # predicted_col = []
            # for j, tile_view in enumerate(tile_view_col):
            label_r = "profile unet cumulative get_region_image time [s]"
            label_p = "profile unet cumulative prediction time [s]"
            t_r = (
                self.report.actual_row[label_r]
                if label_r in self.report.actual_row
                else 0
            )
            t_p = (
                self.report.actual_row[label_p]
                if label_p in self.report.actual_row
                else 0
            )
            logger.trace(
                f"predicting tile {i}, loc={tile_view.region_location}, sz={tile_view.region_size_on_level}, t_r={t_r}, t_p={t_p}"
            )
            # self._inner_texture.texture_label = f"slide_segmentation_{i},{j}"
            predicted_image = self.predict_on_view(tile_view)
            # if str(self.parameters.param("Segmentation Method").value()) == "U-Net":
            #     predicted_image = self._unet.predict_tile(tile_view)
            # elif str(self.parameters.param("Segmentation Method").value()) == "HCTFS":
            #     predicted_image = self.predict_on_view_hctfs(tile_view)
            # else:
            #     raise ValueError("Unknown segmentation method")
            self.predicted_tiles.append(predicted_image)

    def predict(self):
        """
        predict tiles and compose everything together
        """
        logger.debug("predict")
        # if self.predicted_tiles is None:
        #     self.predict_tiles()
        if str(self.parameters.param("Segmentation Method").value()) == "U-Net":
            self._unet.init_segmentation()

        if self.predicted_tiles is None:
            self.predict_tiles()

        # szx = len(self.tiles)
        # szy = len(self.tiles[0])
        #         print(f"size x={szx} y={szy}")

        (
            imsize,
            tile_size_on_level0,
            tile_size_on_level,
            imsize_on_level,
        ) = self._get_tiles_parameters()
        output_image = np.zeros(self.tile_size + np.asarray(imsize_on_level), dtype=int)
        logger.debug("composing predicted image")
        # for iy, tile_column in enumerate(self.tiles):
        # for ix, tile in enumerate(tile_column):
        for (view, tile_params), predicted_tile in zip(
            self.tiles, self.predicted_tiles
        ):
            sl_tl, sl_gl, loc = tile_params
            output_image[
                # ix * self.tile_size[0] : (ix + 1) * self.tile_size[0],
                # iy * self.tile_size[1] : (iy + 1) * self.tile_size[1],
                sl_gl
            ] = predicted_tile

        # full_image = output_image[: int(imsize_on_level[0]), : int(imsize_on_level[1])]

        logger.debug(f"output_image.shape={output_image.shape}")
        full_image = output_image[self._full_image_crop_slice()]
        self.full_prefilter_image = full_image
        # if str(self.parameters.param("Segmentation Method").value()) in ["HCTFS", "GLCMTFS"]:
        if bool(self.parameters.param("Postprocessing").value()):
            self.full_output_image = self._labeling_filtration(full_image)
        else:
            self.full_output_image = full_image
        return self.full_output_image

    def _full_image_crop_slice(self):
        # (
        #     imsize,
        #     tile_size_on_level0,
        #     tile_size_on_level,
        #     imsize_on_level,
        # ) = self._get_tiles_parameters()
        # self._cr_sz_on_level
        logger.debug(f"cr size on level = {self._cr_sz_on_level}")
        return (
            slice(None, self._cr_sz_on_level[0]),
            slice(None, self._cr_sz_on_level[1]),
        )

    def _labeling_filtration(self, full_image):
        """
        smooth label 0 and label 1, keep label 2
        """
        logger.debug("labeling filtration")
        # r_m = float(self.parameters.param("Annotation Radius").value())  # * 1000 # mm
        resolution_m = float(
            self.parameters.param("Working Resolution").value()
        )  # * 1000
        max_dist_m = float(
            self.parameters.param("Postprocessing", "Max. Distance").value()
        )  # * 1000
        tmp_img = full_image.copy()
        tmp_img[full_image == 2] = 1
        import skimage.filters
        mask = scipy.ndimage.distance_transform_edt(full_image != 2, sampling=resolution_m) < max_dist_m
        tmp_img[~mask] = 0

        tmp_img = skimage.filters.gaussian(tmp_img.astype(np.float), sigma=4, preserve_range=False)

        tmp_img = (tmp_img > 0.5).astype(np.int)
        tmp_img[full_image == 2] = 2
        return tmp_img

    def get_raster_image(self, as_gray=False):
        if self.tiles is None:
            self.make_tiles()
        # szx = len(self.tiles)
        # szy = len(self.tiles[0])
        #         print(f"size x={szx} y={szy}")

        _, _, _, imsize_on_level = self._get_tiles_parameters()
        output_size = imsize_on_level + self.tile_size
        if not as_gray:
            output_size = np.asarray([output_size[0], output_size[1], 3])

        # (
        #     imsize,
        #     tile_size_on_level0,
        #     tile_size_on_level,
        #     imsize_on_level,
        # ) = self._get_tiles_parameters()
        output_image = np.zeros(output_size, dtype=int)
        # for iy, tile_column in enumerate(self.tiles):
        #     for ix, tile in enumerate(tile_column):
        #         output_image[
        #             ix * self.tile_size[0] : (ix + 1) * self.tile_size[0],
        #             iy * self.tile_size[1] : (iy + 1) * self.tile_size[1]
        #         ] = self.tiles[iy][ix].get_region_image(as_gray=as_gray)[:, :, :3]
        # for sl_x_tl, sl_y_tl, sl_x_gl, sl_y_gl, iy, ix  in self.tile_iterator(return_tile_coords=True):
        #     output_image[sl_x_gl, sl_y_gl] = self.tiles[iy][ix].get_region_image(as_gray=as_gray)[sl_x_tl, sl_y_tl, :3]

        for view, tile_params in self.tiles:
            (sl_x_tl, sl_y_tl), sl_gl, loc = tile_params

            output_image[sl_gl] = view.get_region_image(as_gray=as_gray)[
                sl_x_tl, sl_y_tl, :3
            ]

        full_image = output_image[self._full_image_crop_slice()]
        self.full_raster_image = full_image
        return full_image

    def tile_iterator(
        self,
        return_in_out_coords=True,
        return_level0_coords=False,
        return_tile_coords=False,
    ):
        (
            imsize_on_level0,
            tile_size_on_level0,
            size_on_level,
            imsize_on_level,
        ) = self._get_tiles_parameters()
        # sp_height0, sp_width0 = self.view.get_size_on_level(0)
        # st_height0, st_width0 = self.view.region_location
        st_height0, st_width0 = self._cr_st_on_level0
        # height0 = sp_height0 - st_height0
        # width0 = sp_width0 - st_width0
        # strange behavior is given by openslide. It swaps x and y. The read_region([y,x]).
        for ix, x0 in enumerate(
            range(
                st_height0, int(self._cr_sp_on_level0[0]), int(tile_size_on_level0[0])
            )
        ):
            for iy, y0 in enumerate(
                range(
                    st_width0,
                    int(self._cr_sp_on_level0[1]),
                    int(tile_size_on_level0[1]),
                )
            ):
                # for ix, x0 in enumerate(range(st_height0, int(imsize_on_level0[0]), int(size_on_level0[0]))):
                #     for iy, y0 in enumerate(range(st_width0, int(imsize_on_level0[1]), int(size_on_level0[1]))):
                # logger.trace(f"processing tile {x0}, {y0}")
                x_start = ix * self.tile_size[0]
                x_stop = (ix + 1) * self.tile_size[0]
                y_start = iy * self.tile_size[1]
                y_stop = (iy + 1) * self.tile_size[1]
                sl_x_gl = slice(x_start, x_stop)
                sl_y_gl = slice(y_start, y_stop)
                sl_x_tl = slice(None, None)
                sl_y_tl = slice(None, None)

                if self.compatible_with_openslide:
                    yield (sl_x_tl, sl_y_tl), (sl_x_gl, sl_y_gl), (
                        y0,
                        x0,
                    )  # hamamatsu ok, zeiss wrong
                else:
                    yield (sl_x_tl, sl_y_tl), (sl_x_gl, sl_y_gl), (x0, y0)
                # if return_in_out_coords:
                #     out.extend()
                # if return_tile_coords:
                #     out.append(ix)
                #     out.append(iy)
                # if return_level0_coords:
                #     out.append(x0)
                #     out.append(y0)
                # yield tuple(out)

                #                     int(x0):int(x0 + tile_size_on_level[0]),
                #                     int(y0):int(y0 + tile_size_on_level[1])
                #                 ] = self.tiles[ix][iy].get_region_image(as_gray=True)
                # ] = self.tiles[iy][ix].get_region_image(as_gray=as_gray)[:, :, :3]

    def evaluate(self):
        logger.debug("evaluate")
        labels, count = np.unique(self.full_output_image, return_counts=True)
        logger.debug(f"whole scan segmentation: labels={labels}, count={count}")
        countd = {0: 0, 1: 0, 2: 0}
        countd.update(dict(zip(labels, count)))
        #         plt.figure(figsize=(10, 10))
        #         plt.imshow(self.full_output_image)
        logger.trace("before imsave slice_label.png")
        self.report.imsave(
            "slice_label.png", self.full_output_image, level_skimage=20, level_npz=30
        )
        self.report.imsave(
            "slice_prefilter_label.png",
            self.full_prefilter_image,
            level=40,
            level_skimage=20,
            level_npz=30,
        )
        # plt.imsave(self.output_label_fn, self.full_output_image)

        #         plt.figure(figsize=(10, 10))
        img = self.get_raster_image(as_gray=False)
        #         plt.imshow(img)
        self.report.imsave(
            "slice_raster.png", img.astype(np.uint8), level_skimage=40, level_npz=30
        )
        fig = plt.figure()
        # plt.imshow(img)
        # view_tmp = self.view
        # the view is wrong if it is based on whole scan image
        view_tmp = self.view.to_pixelsize(self.used_pixelsize_mm)
        plt.imshow(view_tmp.get_raster_image(as_gray=False))

        # logger.debug(f"slice_raster size on pixelsize {self.view.region_pixelsize}, {self.view.annotations}")
        logger.debug(self.view.region_size_on_level)
        logger.debug(f"view.region_level={self.view.region_level}")
        view_tmp.plot_annotations(None)
        # self.view.region_imshow_annotation(i=None)
        # self.view.imshow()
        self.report.savefig("slice_raster_with_annotation.png", level=45)
        plt.close(fig)
        # fig.ax
        self.intralobular_ratio = countd[1] / (countd[1] + countd[2])
        logger.debug(f"real_pixel_size={self.used_pixelsize_mm}")
        self.empty_area_mm = np.prod(self.used_pixelsize_mm) * countd[0]
        self.sinusoidal_area_mm = np.prod(self.used_pixelsize_mm) * countd[1]
        self.septum_area_mm = np.prod(self.used_pixelsize_mm) * countd[2]
        logger.debug(f"empty_area_mm={self.empty_area_mm}")
        self.report.set_persistent_cols(
            {
                "Scan Segmentation Empty Area [mm^2]": self.empty_area_mm,
                "Scan Segmentation Septum Area [mm^2]": self.septum_area_mm,
                "Scan Segmentation Sinusoidal Area [mm^2]": self.sinusoidal_area_mm,
                "Scan Segmentation Used Pixelsize [mm]": self.used_pixelsize_mm[0],
                "Scan Segmentation Classifier": str(self.clf),
            }
        )

    def evaluate_labels(self):

        if self.whole_slide_training_labels is not None and self.full_output_image is not None:
            # labels_true = (self.whole_slide_training_labels - 1).astype(np.int8)

            selection = self.whole_slide_training_labels > 0
            if np.sum(selection) > 0:
                accuracy = sklearn.metrics.accuracy_score(
                    self.whole_slide_training_labels[selection].ravel() - 1,
                    self.full_output_image[selection].ravel(),
                    normalize=True
                )
                self.report.set_persistent_cols(
                    {
                        "Whole Scan Training Labels Accuracy": accuracy
                    }
                )
                self.accuracy = accuracy
                fig = plt.figure()
                plt.imshow(self.full_output_image, cmap="Purples")
                plt.contour(self.whole_slide_training_labels, cmap="Reds")
                self.report.savefig("segmentation_and_training_labels.png", level=45)
                plt.close(fig)
                return accuracy
        return None


    def _find_biggest_lobuli(self):
        """
        :param n_max: Number of points. All points are returned if set to negative values.
        The minimum distance between two maximums is given by Annotation Radius parameter (2*annotation_radius).
        """
        n_max = int(self.parameters.param("Lobulus Number").value())
        mask = self.full_output_image == 1

        import time

        # t0 = time.time()
        dist = scipy.ndimage.morphology.distance_transform_edt(mask)
        lab = skimage.morphology.label(mask)

        self.dist = dist
        # report
        r_m = float(self.parameters.param("Annotation Radius").value())  # * 1000 # mm
        resolution_m = float(
            self.parameters.param("Working Resolution").value()
        )  # * 1000

        min_distance = int(2 * r_m / resolution_m)
        logger.debug(f"minimum distance [px]: {min_distance}")

        # image_max = scipy.ndimage.maximum_filter(dist, size=min_distance, mode="constant")
        # Comparison between image_max and im to find the coordinates of local maxima
        coordinates = peak_local_max(dist, min_distance=min_distance)
        point_dist = dist[tuple(list(zip(*coordinates)))]
        # display(point_dist)
        # max_point_inds = point_dist.argsort()[-n_max:][::-1]
        sorted_point_inds = point_dist.argsort()[::-1]
        sorted_points = coordinates[sorted_point_inds]
        centers_filtered = kick_close_points(sorted_points, min_distance=min_distance)

        max_points = centers_filtered[: min(n_max, len(centers_filtered))]
        self.centers_all = coordinates
        self.centers_max = max_points

        # t1 = time.time()
        # ta = t1 - t0
        # # alternative calculation
        #
        # dists, coords = find_maxdist_in_labeled_image(mask)
        # coords_max = coords[:n_max]
        # t2 = time.time()
        # tb = t2 - t1
        # coordinates=coords
        # max_points = coords_max

        #     report
        fig = plt.figure(figsize=(10, 10))
        plt.imshow(dist, cmap=plt.cm.gray)
        plt.autoscale(False)
        plt.plot(
            coordinates[:, 1],
            coordinates[:, 0],
            "g.",
            max_points[:, 1],
            max_points[:, 0],
            "ro",
        )
        plt.axis("off")
        self.report.savefig(
            "sinusoidal_tissue_local_centers.png", level=55
        )
        self.report.savefig_and_show(
            "sinusoidal_tissue_local_centers.pdf", fig, level=55
        )

        return max_points

    def add_biggest_to_annotations(self):
        points_px = self._find_biggest_lobuli()
        # view_corner = self.tiles[0][0]
        view_corner, _ = self.tiles[0]
        pts_glob_px = view_corner.coords_view_px_to_glob_px(
            points_px[:, 1], points_px[:, 0]
        )
        centers_px = list(zip(*pts_glob_px))
        r_mm = float(self.parameters.param("Annotation Radius").value()) * 1000
        biggest_ids, _ = add_circle_annotation(
            view_corner, centers_px, r_mm, self.anim.annotations
        )

        # self.anim.annotations.extend(anns)
        self.ann_biggest_ids.extend(biggest_ids)

    def _prepare_clf_fn(self):
        method = str(self.parameters.param("Segmentation Method").value())

    def run(self):
        logger.debug("run...")
        # GLCM
        # self._inner_texture.set_report(self.report)
        # self._inner_texture.add_cols_to_report = False

        method = str(self.parameters.param("Segmentation Method").value())
        if method in ("HCTFS", "GLCMTFS"):
            if bool(self.parameters.param("TFS General", "Load Default Classifier").value()):
                if self.clf_default_fn.exists():
                    logger.debug(
                        f"Reading default classifier from {str(self.clf_default_fn)}"
                    )
                    self.clf = joblib.load(self.clf_default_fn)
                else:
                    logger.error("Default classifier not found")

        if bool(self.parameters.param("TFS General", "Run Training").value()):
            self.train_classifier()
            self.save_classifier()
        if bool(self.parameters.param("Save Training Labels").value()):
            self.save_training_labels()
        if bool(self.parameters.param("Run Prediction").value()):
            logger.debug("predict...")
            self.predict()
            logger.debug("evaluate...")
            self.evaluate()
            self.evaluate_labels()
            self._imsave_full_raster_and_training_labels()

        res = list(self.anim.get_pixel_size(self.level))
        logger.debug(f"slide segmentation resolution = {res}")
        self.report.set_persistent_cols(
            {
                "slide segmentation method": method,
                "slide segmentation resolution": str(res),
            }
        )


# def find_maxdist_in_labeled_image(mask:np.ndarray):
#     """
#     Find biggest labeled areas in sense of distance from the border
#     :param lab:
#     :return: dists, coords
#     """
#
#     lab = skimage.morphology.label(mask)
#     dist = scipy.ndimage.morphology.distance_transform_edt(mask)
#     mxs = [None] * (np.max(lab)-1)
#     for l in range(1, np.max(lab)):
#         dist_i = dist.copy()
#         dist_i[lab != l] = 0
#         # if l == 0:
#         #     mxs[l] = (l, ())
#         #     continue
#         # dist_i = scipy.ndimage.morphology.distance_transform_edt(lab == l)
#         mx = np.max(dist_i)
#         xyz = np.unravel_index(np.argmax(dist_i), shape=lab.shape)
#         mxs[l-1] = (mx, xyz)
#
#
#     mxs.sort(key=lambda x:x[0], reverse=True)
#
#     dists, coords = zip(*mxs)
#     dists = np.asarray(dists)
#     coords = np.asarray(coords)
#     return dists, coords


def kick_close_points(coords, min_distance):
    selected = []
    for i in range(coords.shape[0] - 1):
        dists = np.linalg.norm((coords[i + 1 :, :] - coords[i, :]), axis=1)
        # print(dists)
        if (dists > min_distance).all():
            selected.append(coords[i, :])
        else:
            logger.trace(f"kicked {coords[i, :]}")

    selected = np.asarray(selected)
    return selected


def add_circle_annotation(view_corner: scim.View, centers_px_global, r_mm, annotations):

    # r_mm = 0.1
    t = np.linspace(0, 2 * np.pi, 30)

    anns = []
    biggest_ids = []
    logger.debug(f"Automatic selection centers_px={centers_px_global}")
    for center_px in centers_px_global:
        r_px = view_corner.mm_to_px(r_mm)
        #     print(f"r_px={r_px}")
        r_px_glob = view_corner.coords_view_px_to_glob_px(
            np.array([r_px[0]]), np.array([r_px[1]])
        )
        x_px = r_px_glob[0] * np.sin(t) + center_px[0]
        y_px = r_px_glob[1] * np.cos(t) + center_px[1]

        ann = {
            "title": "Automatic Selection",
            "x_px": x_px,
            "y_px": y_px,
            "color": "#00FF88",
            "details": "",
        }
        ann = annotation_px_to_mm(view_corner.anim.openslide, ann)
        newid = len(annotations)
        annotations.append(ann)
        biggest_ids.append(newid)

    return biggest_ids, annotations
    # self.ann_biggest_ids = new_ann_ids



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Modul is used for texrure analysis.
"""

from loguru import logger

# import warnings
import numpy as np
import scipy.ndimage
import matplotlib.pyplot as plt
import copy
from typing import List
import os.path as op
from pyqtgraph.parametertree import Parameter, ParameterTree
from . import image
from exsu.report import Report
import imma.image
from . import sni_prediction


def tile_centers(image_shape, tile_spacing):
    tile_size2 = [int(tile_spacing[0] / 2), int(tile_spacing[1] / 2)]
    centers = []
    for x0 in range(0, image_shape[0], tile_spacing[0]):
        for x1 in range(0, image_shape[1], tile_spacing[1]):
            centers.append([x0 + tile_size2[0], x1 + tile_size2[1]])
    return centers


def tiles_processing(
    image, fcn, tile_spacing, fcn_output_n=None, dtype=np.int8, tile_size=None
):
    """
    Process image tile by tile. Last tile in every row and avery column may be smaller if modulo of shape of image and
    shape of tile is different from zero. On the border of image tile is smaller according to the edge of the image.

    :param image: input image
    :param fcn: Function used on each tile. Input of this function is just tile image.
    :param tile_spacing: size of tile in pixels
    :param fcn_output_n: dimension of output of fcn()
    :param dtype: output data type
    :param tile_size: [int, int]: size of tile in pixels. Tile size can be set to obtain overlap between two
    neighbour tiles.
    :return:
    """
    # TODO rename inputs

    shape = list(image.shape)
    if fcn_output_n is not None:
        shape.append(fcn_output_n)
    if tile_size is None:
        tile_margin = [0, 0]
    else:
        tile_margin = ((np.asarray(tile_size) - tile_spacing) / 2).astype(np.int)

    output = np.zeros(shape, dtype=dtype)
    for x0 in range(0, image.shape[0], tile_spacing[0]):
        for x1 in range(0, image.shape[1], tile_spacing[1]):
            sl_in = (
                slice(
                    max(x0 - tile_margin[0], 0), x0 + tile_spacing[0] + tile_margin[0]
                ),
                slice(
                    max(x1 - tile_margin[1], 0), x1 + tile_spacing[1] + tile_margin[1]
                ),
            )
            sl_out = (slice(x0, x0 + tile_spacing[0]), slice(x1, x1 + tile_spacing[1]))
            img = image[sl_in]
            output[sl_out] = fcn(img)

    else:
        return output


def get_feature_and_predict(img, fv_function, classif):
    fv = fv_function(img)
    return classif.predict([fv])[0]


def select_texture_patch_centers_from_one_annotation(
    anim, title, tile_size, level, step=50
):
    if not np.isscalar(tile_size):
        if tile_size[0] == tile_size[1]:
            tile_size = tile_size[0]
        else:
            # it would be possible to add factor (1./tile_size) into distance transform
            raise ValueError(
                "Both sides of tile should be the same. Other option is not implemented."
            )
    annotation_ids = anim.select_annotations_by_title(title)
    view = anim.get_views(annotation_ids, level=level)[0]
    mask = view.get_annotation_raster(title)

    # with warnings.catch_warnings():
    #     warnings.filterwarnings("ignore", "low contrast image")
    dst = scipy.ndimage.morphology.distance_transform_edt(mask)
    middle_pixels = dst > (tile_size / 2)
    # x_nz, y_nz = nonzero_with_step(middle_pixels, step)
    y_nz, x_nz = nonzero_with_step(middle_pixels, step)
    nz_global_px = view.coords_view_px_to_glob_px(x_nz, y_nz)
    # anim.
    return nz_global_px


def nonzero_with_step(data, step):
    # print(data.shape)
    datastep = data[::step, ::step]
    # print(datastep.shape)
    nzx, nzy = np.nonzero(datastep)

    return nzx * step, nzy * step


class GLCMTextureMeasurement:
    def __init__(
        self,
        filename_label="",
        pname="Texture Analysis",
        ptype="bool",
        pvalue=True,
        ptip="Run Gray Level Co-occurrence texture analysis after lobulus segmentation is performed",
        texture_label=None,
        working_resolution_mm=0.0000004,
        tile_size=64,
        tile_spacing=32,
        add_cols_to_report: bool = True,
        report_severity_offset=0,
        glcm_levels=64,
        report: Report = None,
        sni_predictor=None,
    ):
        """
        Make texture description.
        It is used twice in scaffan.

        :param filename_label:
        :param pname:
        :param ptype:
        :param pvalue:
        :param ptip:
        :param texture_label: Used to identify where the texture processing is used (central vein, whole_slide, ...)
        :param working_resolution_mm:
        :param tile_size:
        :param tile_spacing:
        """
        self.sni_predictor = sni_predictor
        params = [
            {
                "name": "Tile Size",
                "type": "int",
                # "value": 64,
                "value": tile_size,
                "suffix": "px",
            },
            {
                "name": "Tile Spacing",
                "type": "int",
                # "value": 64,
                "value": tile_spacing,
                "suffix": "px",
            },
            {
                "name": "Working Resolution",
                "type": "float",
                # "value": 0.000001,
                # "value": 0.0000005,
                # "value": 0.0000004,
                "value": working_resolution_mm,
                "suffix": "m",
                "siPrefix": True,
            },
            {"name": "GLCM Levels", "type": "int", "value": glcm_levels},
        ]
        if self.sni_predictor is not None:
            params.append(self.sni_predictor.parameters)
        self.texture_label = texture_label

        self.parameters = Parameter.create(
            name=pname,
            type=ptype,
            value=pvalue,
            tip=ptip,
            children=params,
            expanded=False,
        )
        self.report: Report = report
        self.filename_label = filename_label
        self.add_cols_to_report: bool = add_cols_to_report
        self.annotation_id = None
        self.lobulus_segmentation = None
        self.report_severity_offset = report_severity_offset

    def set_report(self, report: Report):
        self.report = report

    def set_input_data(
        self, view: image.View, annotation_id:int=None, lobulus_segmentation=None,
            texture_label_str_id:str=None
    ):
        self.anim = view.anim
        self.annotation_id = annotation_id
        self.parent_view = view
        logger.trace(f"lobulus segmentation {lobulus_segmentation}")
        self.lobulus_segmentation = lobulus_segmentation
        self.texture_label_str_id = texture_label_str_id

    def run(self, recalculate_view=True):
        """

        :param recalculate_view: if False do not calculate other view paramters and use this one
        :return:
        """

        # title = "test3"
        # title = "test2"
        # title = "test1"
        # views = self.anim.get_views_by_title(self.annotation_id, level=0)
        pxsize_mm = [self.parameters.param("Working Resolution").value() * 1000] * 2
        tile_size = [self.parameters.param("Tile Size").value()] * 2
        tile_spacing = [self.parameters.param("Tile Spacing").value()] * 2
        levels = self.parameters.param("GLCM Levels").value()
        # view = views[0].to_pixelsize(pxsize_mm)
        if self.texture_label is not None:
            texture_label_fn = "_" + self.texture_label
        else:
            texture_label_fn = ""
        if recalculate_view:
            view = self.parent_view.to_pixelsize(pxsize_mm)
        else:
            view = self.parent_view

        texture_image = view.get_region_image(as_gray=True)

        fn_ann_id = "" if self.annotation_id is None else f"_{self.annotation_id}"
        tx_ann_id = "" if self.texture_label_str_id is None else f"_{self.texture_label_str_id}"
        fn_id = "{}{}{}{}".format(self.filename_label, fn_ann_id, texture_label_fn, tx_ann_id)
        if self.report is not None:
            self.report.imsave(
                f"texture_input_image_mul255_{fn_id}.png",
                (texture_image * 255).astype(np.uint8),
                level=45 + self.report_severity_offset,
                level_npz=40 + self.report_severity_offset,
                level_skimage=20 + self.report_severity_offset,
            )
            self.report.imsave(
                f"texture_input_image_mul1_{fn_id}.png",
                texture_image,
                level=45 + self.report_severity_offset,
                level_npz=40 + self.report_severity_offset,
                level_skimage=20 + self.report_severity_offset,
            )
        energy = tiles_processing(
            1 - texture_image,
            fcn=lambda img: texture_glcm_features(img, levels),
            tile_spacing=tile_spacing,
            fcn_output_n=3,
            dtype=None,
            tile_size=tile_size,
        )
        # seg = texseg.predict(views[0], show=False, function=texture_energy)
        fig = plt.figure(figsize=(10, 12))
        plt.subplot(221)
        # grayscale image is there because of travis memory error
        img = view.get_region_image(as_gray=True)
        plt.imshow(img)
        # if self.annotation_id is not None:
        view.plot_annotations(self.annotation_id)
        if self.lobulus_segmentation is not None:
            seg = (
                imma.image.resize_to_shape(
                    self.lobulus_segmentation, shape=img.shape[:2], order=0
                )
                == 1
            )
            plt.contour(seg)
            plt.title("original image")
        plt.subplot(222)
        plt.title("GLCM energy")
        image.imshow_with_colorbar(energy[:, :, 0])
        plt.subplot(223)
        plt.title("GLCM homogeneity")
        image.imshow_with_colorbar(energy[:, :, 1])
        plt.subplot(224)
        plt.title("GLCM correlation")
        image.imshow_with_colorbar(energy[:, :, 2])
        mx = np.max(energy, axis=(0, 1))
        mn = np.min(energy, axis=(0, 1))
        logger.debug(f"GLCM max energy, homogeneity, correlation: {mx}")
        logger.debug(f"GLCM min energy, homogeneity, correlation: {mn}")
        # plt.colorbar()
        if self.report is not None:
            self.report.savefig_and_show(
                "glcm_features_{}.png".format(fn_id),
                fig,
                level=60 + self.report_severity_offset,
            )
        # if close_figs:
        #     plt.close(fig)
        # plt.savefig("glcm_features_{}.png".format(title))
        logger.debug(f"glcm_features_{fn_id} saved")
        fig = plt.figure()
        # 0..1 normalization because energy (3rd channel) can be negative
        w = np.ones([1, 1, 3])
        w.ravel()[2] = 0.5
        t = np.zeros([1, 1, 3])
        t.ravel()[2] = 0.5
        plt.imshow(energy * w + t)
        if self.report is not None:

            self.report.savefig_and_show(
                "glcm_features_color_{}.png".format(fn_id),
                fig,
                level=60 + self.report_severity_offset,
            )
        # if close_figs:
        #     plt.close(fig)

        e0 = energy[:, :, 0]
        e1 = energy[:, :, 1]
        e2 = energy[:, :, 2]
        self.measured_features = energy

        if self.lobulus_segmentation is None:
            logger.debug(f"No lobulus segmentation given")
            seg = (slice(None), slice(None))
        if self.texture_label is None:
            texture_label_stats = ""
        else:
            texture_label_stats = f" {self.texture_label}"

        row = {}
        row = make_stats(f"GLCM Energy{texture_label_stats}", e0[seg], row)
        row = make_stats(f"GLCM Homogenity{texture_label_stats}", e1[seg], row)
        row = make_stats(f"GLCM Correlation{texture_label_stats}", e2[seg], row)
        # row = {
        #     "GLCM Energy": np.mean(e0[seg]),
        #     "GLCM Homogenity": np.mean(e1[seg]),
        #     "GLCM Correlation": np.mean(e2[seg]),
        # }
        if self.report:
            self.report.imsave_as_fig(
                f"glcm_texture_features_{fn_id}",
                # (self.measured_features * 255).astype(np.uint8),
                self.measured_features,
                level=45 + self.report_severity_offset,
                # level_npz=55 + self.report_severity_offset,
                npz_level=40 + self.report_severity_offset,
                # level_skimage=0,
                # level_skimage=45 + self.report_severity_offset,
            )
        if self.sni_predictor is not None:
            # texture processing is called twice
            self.sni_predictor.predict_area(row)
        if self.report is not None and self.add_cols_to_report:
            self.report.add_cols_to_actual_row(row)

        logger.debug(
            f"GLCM textures for id={self.annotation_id if self.annotation_id is not None else '-'} "
            f"(tx_label={tx_ann_id}) finished"
        )

        # plt.show()


def make_stats(prefix: str, data, dct=None):
    """
    Calculate mean, variance and 10, 25, 75 and 90 percentiles. Output
    is stored to dict with key starting with `prefix`
    :param prefix: prefix of output dict key
    :param data: numeric data to analyze
    :param dct: input dict. Empty dict is created if None is given.
    :return:
    """
    if dct is None:
        dct = {}

    dct[prefix] = np.mean(data)
    dct[f"{prefix} var"] = np.var(data)
    dct[f"{prefix} p10"] = np.percentile(data, 10)
    dct[f"{prefix} p25"] = np.percentile(data, 25)
    dct[f"{prefix} p50"] = np.percentile(data, 50)
    dct[f"{prefix} p75"] = np.percentile(data, 75)
    dct[f"{prefix} p90"] = np.percentile(data, 90)
    return dct


class TextureSegmentation:
    def __init__(self, feature_function=None, classifier=None):

        params = [
            {"name": "Tile Size", "type": "int", "value": 256},
            # {
            #     "name": "Working Resolution",
            #     "type": "float",
            #     "value": 0.001,
            #     "suffix": "m",
            #     "siPrefix": True
            #
            # }
        ]

        self.parameters = Parameter.create(
            name="Texture Processing", type="group", children=params
        )
        self.parameters
        self.tile_size = None
        self.tile_size1 = None
        self.set_tile_size(self.parameters.param("Tile Size").value())
        self.parameters.param("Tile Size").sigValueChanged.connect(
            self._seg_tile_size_params
        )

        self.level = 1
        self.step = 64
        self.data = []
        self.target = []
        if feature_function is None:
            import scaffan.texture_lbp as salbp

            feature_function = salbp.lbp_fv
        self.feature_function = feature_function
        if classifier is None:
            import scaffan.texture_lbp as salbp

            classifier = salbp.KLDClassifier()
        self.classifier = classifier

        # n_points = 8
        # radius = 3
        # METHOD = "uniform"
        # self.feature_function_args = [n_points, radius, METHOD]
        logger.debug("texture run finished")

    def _seg_tile_size_params(self):
        self.set_tile_size(self.parameters.param("Tile Size").value())

    def set_tile_size(self, tile_size1):
        self.tile_size = [tile_size1, tile_size1]
        self.tile_size1 = tile_size1

    def get_tile_centers(self, anim, annotation_id, return_xy=False):
        """
        Calculate centers for specific annotation.
        :param anim:
        :param annotation_id:
        :return: [[x0, y0], [x1, y1], ...]
        """
        patch_centers1 = select_texture_patch_centers_from_one_annotation(
            anim,
            annotation_id,
            tile_size=self.tile_size1,
            level=self.level,
            step=self.step,
        )
        if return_xy:
            return patch_centers1
        else:
            patch_centers1_points = list(zip(*patch_centers1))
            return patch_centers1_points

    def get_patch_view(
        self, anim: image.AnnotatedImage, patch_center=None, annotation_id=None
    ):
        if patch_center is not None:
            view: image.View = anim.get_view(
                center=[patch_center[0], patch_center[1]],
                level=self.level,
                size_on_level=self.tile_size,
            )
        elif patch_center is not None:
            annotation_ids = anim.select_annotations_by_title(
                title=annotation_id, level=self.level, size=self.tile_size
            )
            view = anim.get_views(annotation_ids)[0]

        return view

    def show_tiles(self, anim, annotation_id, tile_ids):
        """
        Show tiles from annotation selected by list of its id
        :param anim:
        :param annotation_id:
        :param tile_ids: list of int, [0, 5] means show first and sixth tile
        :return:
        """
        patch_center_points = self.get_tile_centers(anim, annotation_id)
        for id in tile_ids:
            view = self.get_patch_view(anim, patch_center_points[id])
            plt.figure()
            plt.imshow(view.get_region_image(as_gray=True), cmap="gray")

    def add_training_data(self, anim, annotation_id, numeric_label, show=False):
        patch_center_points = self.get_tile_centers(anim, annotation_id)

        for patch_center in patch_center_points:
            view = self.get_patch_view(anim, patch_center)
            imgray = view.get_region_image(as_gray=True)
            self.data.append(self.feature_function(imgray))
            self.target.append(numeric_label)

        if show:
            annotation_ids = anim.select_annotations_by_title(title=annotation_id)
            view = anim.get_views(annotation_ids)[0]
            view.imshow()
            lst = list(zip(*patch_center_points))
            x, y = lst
            view.plot_points(x, y)
        return patch_center_points

    def fit(self):
        self.classifier.fit(self.data, self.target)
        pass

    def predict(self, view, show=False):
        test_image = view.get_region_image(as_gray=True)

        tile_fcn = lambda img: get_feature_and_predict(
            img, self.feature_function, self.classifier
        )
        seg = tiles_processing(test_image, tile_fcn, tile_spacing=self.tile_size)

        if show:
            centers = tile_centers(test_image.shape, tile_spacing=self.tile_size)
            import skimage.color

            plt.imshow(skimage.color.label2rgb(seg, test_image))
            x, y = list(zip(*centers))
            plt.plot(x, y, "xy")
            # view.plot_points()
        return seg


def texture_glcm_features(img, levels):
    import skimage.feature.texture

    # levels =
    # if distances is None:
    distances = [1]
    # if angles is None:
    angles = [0, np.pi / 2]

    P = skimage.feature.greycomatrix(
        (img * (levels - 1)).astype(np.uint8),
        distances,
        angles,
        levels=levels,
        symmetric=True,
        normed=True,
    )
    en = skimage.feature.texture.greycoprops(P, prop="energy")
    # dissimilarity = skimage.feature.texture.greycoprops(P, prop="dissimilarity")
    homogeneity = skimage.feature.texture.greycoprops(P, prop="homogeneity")
    correlation = skimage.feature.texture.greycoprops(P, prop="correlation")
    return np.array([np.mean(en), np.mean(homogeneity), np.mean(correlation)])


def texture_energy(img):
    import skimage.feature.texture

    P = skimage.feature.greycomatrix(
        (img * 31).astype(np.uint8),
        [1],
        [0, np.pi / 2],
        levels=32,
        symmetric=True,
        normed=True,
    )
    en = skimage.feature.texture.greycoprops(P, prop="energy")
    return np.mean(en) * 100



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Modul is used for texrure analysis.
"""

from loguru import logger
import numpy as np
import skimage.feature
import matplotlib.pyplot as plt

# settings for LBP
radius = 2
n_points = 8 * radius
method = "uniform"


def kullback_leibler_divergence(p, q):
    p = np.asarray(p)
    q = np.asarray(q)
    filt = np.logical_and(p != 0, q != 0)
    return np.sum(p[filt] * np.log2(p[filt] / q[filt]))


def lbp_fv(img):  # , n_points, radius, METHOD):
    """

    :param img: img in range 0.0-1.0
    :return:
    """
    img = (img * 255).astype(np.uint8)
    lbp = skimage.feature.local_binary_pattern(img, n_points, radius, method)
    n_bins = int(lbp.max() + 1)
    hist, _ = np.histogram(lbp, density=True, bins=n_bins, range=(0, n_bins))
    return hist


# def match(refs, img):
#     best_score = 10
#     best_name = None
#     if type(refs) is dict:
#         itms = refs.items()
#     elif type(refs) is list:
#         itms = refs
#     else:
#         ValueError("Wrong type for 'refs'")
#
#     for name, ref in itms:
#         ref_hist, _ = np.histogram(ref, density=True, bins=n_bins,
#                                    range=(0, n_bins))
#         score = kullback_leibler_divergence(hist, ref_hist)
#         if score < best_score:
#             best_score = score
#             best_name = name
#     return best_name


def show_lbp(lbp):
    n_bins = int(lbp.max() + 1)
    plt.hist(lbp.ravel(), range=(0, n_bins), bins=n_bins, normed=True)
    plt.xlim(xmax=n_points + 2)


class KLDClassifier:
    def __init__(self):
        pass

    def fit(self, data, target):

        self.data = data
        self.target = target
        # self.refs = list(zip(data, target))
        pass

    def predict_one(self, hist):
        best_score = np.inf
        for name, ref_hist in zip(self.target, self.data):
            score = kullback_leibler_divergence(hist, ref_hist)
            if score < best_score:
                best_score = score
                best_name = name
        return best_name

    def predict(self, x):
        out = [None] * len(x)
        for i, hist in enumerate(x):
            out[i] = self.predict_one(hist)

        return out



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Modul is used for GUI of Lisa
"""

from loguru import logger

from scaffan import image
import sys
import os.path as op
import datetime
from pathlib import Path
import io3d.misc
from io3d import cachefile
import json
import time
import platform
from typing import List, Union, Optional

# import PyQt5.QtWidgets
# print("start 3")
# from PyQt5.QtWidgets import QApplication, QFileDialog
# print("start 4")
from PyQt5 import QtGui

# print("start 5")
from pyqtgraph.parametertree import Parameter, ParameterTree
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

# matplotlib.use('Qt5Agg')

# print("start 6")

import io3d
import io3d.datasets
import scaffan.lobulus
import scaffan.skeleton_analysis
from exsu.report import Report
import scaffan.evaluation
import scaffan.slide_segmentation
from scaffan.pyqt_widgets import BatchFileProcessingParameter
from scaffan.image_intensity_rescale_pyqtgraph import RescaleIntensityPercentilePQG
from . import sni_prediction
from . import lobule_quality_estimation_cnn


class Scaffan:
    def __init__(self, whole_scan_margin=-0.0, default_output_dir_prefix=None):
        """

        :param whole_scan_margin: negative value makes the the area for automatic lobuli selection smaller. It is used in
        tests and debugging.
        :param default_output_dir_prefix: Set prefix dir. Into this dir will be placed output dir with name generated by
        date and time.
        """

        self.report: Report = Report(
            repodir=Path(__file__).parent.resolve(),
            check_version_of=["numpy", "scipy", "skimage"],
        )
        self._default_output_dir_prefix = default_output_dir_prefix if \
            default_output_dir_prefix else self._prepare_default_output_dir_prefix()
        # self.report.level = 50

        self.raise_exception_if_color_not_found = True

        import scaffan.texture as satex

        sni_predictor = sni_prediction.SniPredictor(
            report=self.report, ptype="bool", pvalue=True
        )
        self.glcm_textures = satex.GLCMTextureMeasurement(
            report=self.report, sni_predictor=sni_predictor
        )
        self.lobulus_processing = scaffan.lobulus.Lobulus(
            ptype="bool", report=self.report
        )
        self.whole_scan_margin = whole_scan_margin
        self.skeleton_analysis = scaffan.skeleton_analysis.SkeletonAnalysis()
        self.evaluation = scaffan.evaluation.Evaluation()
        self.intensity_rescale = RescaleIntensityPercentilePQG()
        self.slide_segmentation = scaffan.slide_segmentation.ScanSegmentation(
            report=self.report
        )
        self.lobule_quality_estimation_cnn = lobule_quality_estimation_cnn.LobuleQualityEstimationCNN(
            report=self.report
        )
        # self.slide_segmentation.report = self.report

        # self.lobulus_processing.set_report(self.report)
        # self.glcm_textures.set_report(self.report)
        self.skeleton_analysis.set_report(self.report)
        self.evaluation.report = self.report
        self.win: QtGui.QWidget = None
        self.cache = cachefile.CacheFile("~/.scaffan_cache.yaml")
        # self.cache.update('', path)
        common_spreadsheet_file = self.cache.get_or_save_default(
            "common_spreadsheet_file",
            self._prepare_default_output_common_spreadsheet_file(),
        )
        logger.debug(
            "common_spreadsheet_file loaded as: {}".format(common_spreadsheet_file)
        )
        params = [
            {
                "name": "Input",
                "type": "group",
                "children": [
                    {"name": "File Path", "type": "str",
                         "tip": "Annotations are imported automatically. Black, red and magenta annotations can be used"
                         + "in training as intra-lobular, extra-lobular and background.\n\n"
                         + "The annotations from `.ndpi` files are automativally imported.\n "
                         + "The annotattion for other image format are expected to be done in ImageJ ROI Manager.\n"
                           "File with annotations for is expected to be in the same dir and same name "
                           "with file extension `.roi.zip`. \n"
                           "Color of segmentation should be part of the polyon name ('my annotation #00FF00').\n"
                           "There has to be one rectangle annotataion covering the whole image. It is used to detect"
                           "original size of image used for generating ImageJ ROI."
                     },
                    {"name": "Select", "type": "action"},
                    {"name": "Data Info", "type": "str", "readonly": True},
                    {
                        "name": "Lobulus Selection Method",
                        "type": "list",
                        "value": "Auto",
                        "values": ["Color", "Manual", "Auto"],
                        # "Color": "Color",
                        # "Manual": "Manual",
                        # "Auto": "Auto",
                        # },
                        "tip": "Auto: select lobulus based on Scan Segmentation.\n"
                        "Color: based on annotation color.\n"
                        "Manual: manually pick the lobule. \n\n"
                    },
                    # {
                    #     "name": "Lobulus Selection Method",
                    #     "type": "bool",
                    #     "value": True,
                    #     "tip": "Skip selection based on annotation color and select lobulus based on Scan Segmentation. ",
                    # },
                    {
                        "name": "Annotation Color",
                        "type": "list",
                        "tip": "Select lobulus based on annotation color. "
                        "Color of the annotation from ImageJ is expected to be part of the polyon name ('my annotation #00FF00')"
                        "Skipped if Automatic Lobulus Selection Method is used.",
                        "values": {
                            "None": None,
                            "White": "#FFFFFF",
                            "Black": "#000000",
                            "Red": "#FF0000",
                            "Green": "#00FF00",
                            "Blue": "#0000FF",
                            "Cyan": "#00FFFF",
                            "Magenta": "#FF00FF",
                            "Yellow": "#FFFF00",
                        },
                        "value": 0,
                    },
                    # {'name': 'Boolean', 'type': 'bool', 'value': True, 'tip': "This is a checkbox"},
                    # {'name': 'Color', 'type': 'color', 'value': "FF0", 'tip': "This is a color button"},
                    # BatchFileProcessingParameter(
                    #     name="Batch processing", children=[]
                    # ),
                    # {
                    #     "name": "Advanced",
                    #     "type": "group",
                    #     "children": [
                    #         dict(name="Ignore not found color",type="bool", value=False,
                    #              tip="No exception thrown if color not found in the data"),
                    #     ]
                    # }
                ],
            },
            {
                "name": "Output",
                "type": "group",
                "children": [
                    {
                        "name": "Directory Path",
                        "type": "str",
                        "value": self._prepare_default_output_dir(),
                    },
                    {"name": "Select", "type": "action"},
                    {
                        "name": "Common Spreadsheet File",
                        "type": "str",
                        "value": common_spreadsheet_file,
                    },
                    {
                        "name": "Select Common Spreadsheet File",
                        "type": "action",
                        "tip": "All measurements are appended to this file in addition to data stored in Output Directory Path.",
                    },
                ],
            },
            {
                "name": "Processing",
                "type": "group",
                "children": [
                    # {'name': 'Directory Path', 'type': 'str', 'value': prepare_default_output_dir()},
                    {
                        "name": "Show",
                        "type": "bool",
                        "value": False,
                        "tip": "Show images",
                    },
                    {
                        "name": "Open output dir",
                        "type": "bool",
                        "value": False,
                        "tip": "Open system window with output dir when processing is finished",
                    },
                    # {
                    #     "name": "Run Scan Segmentation",
                    #     "type": "bool",
                    #     "value": True,
                    #     "tip": "Run analysis of whole scan before all other processing is perfomed",
                    # },
                    # {
                    #     "name": "Skeleton Analysis",
                    #     "type": "bool",
                    #     "value": True,
                    #     # "tip": "Show images",
                    # },
                    # {
                    #     "name": "Texture Analysis",
                    #     "type": "bool",
                    #     "value": True,
                    #     # "tip": "Show images",
                    # },
                    self.intensity_rescale.parameters,
                    self.slide_segmentation.parameters,
                    self.lobulus_processing.parameters,
                    self.skeleton_analysis.parameters,
                    self.glcm_textures.parameters,
                    self.lobule_quality_estimation_cnn.parameters,
                    {
                        "name": "Whole Scan Margin",
                        "type": "float",
                        "value": self.whole_scan_margin,
                        "tip": "Negative value will crop the whole scan image",
                    },
                    {
                        "name": "Preview Pixelsize",
                        "type": "float",
                        "value": 0.00001,  # 0.02 mm
                        "suffix": "m",
                        "siPrefix": True,
                        "tip": "Pixelsize used for manual lobuli selection. ",
                    },
                    {
                        "name": "Report Level",
                        "type": "int",
                        "value": 50,
                        "tip": "Control ammount of stored images. 0 - all debug imagess will be stored. "
                        "100 - just important images will be saved.",
                    },
                ],
            },
            {"name": "Run", "type": "action"},
        ]
        self.parameters = Parameter.create(name="params", type="group", children=params)
        # here is everything what should work with or without GUI
        self.parameters.param("Input", "File Path").sigValueChanged.connect(
            self._get_file_info
        )
        self.anim: image.AnnotatedImage = None
        pass

    def select_file_gui(self):
        from PyQt5 import QtWidgets

        default_dir = io3d.datasets.join_path(get_root=True)
        # default_dir = op.expanduser("~/data")
        if not op.exists(default_dir):
            default_dir = op.expanduser("~")

        fn, mask = QtWidgets.QFileDialog.getOpenFileName(
            None,
            "Select Input File",
            directory=default_dir,
            filter="Images (*.ndpi *.czi *.tif *.tiff);;NanoZoomer Digital Pathology Image (*.ndpi);;"
            + "Zeiss Image format for microscopes (*.czi);;Tiff image (*.tiff *.tif)",
        )
        self.set_input_file(fn)

    def set_input_file(self, fn: Union[Path, str]):
        fn = str(fn)
        fnparam = self.parameters.param("Input", "File Path")
        fnparam.setValue(fn)
        logger.debug("Set Input File Path to : {}".format(fn))
        # import pdb; pdb.set_trace()
        # print("ahoj")

    def set_output_dir(self, path: Union[str, Path]=None):
        """
        Set directory for all outputs. The standard
        :param path: if no parameter is given the standard path in ~/data/SA_%Date_%Time is selected
        :return:
        """
        if path is None:
            path = self._prepare_default_output_dir()
        logger.debug(f"output directory path={path}")
        fnparam = self.parameters.param("Output", "Directory Path")
        fnparam.setValue(str(path))

    def set_report_level(self, level: int):
        fnparam = self.parameters.param("Processing", "Report Level")
        fnparam.setValue(level)

    def set_common_spreadsheet_file(self, path):
        fnparam = self.parameters.param("Output", "Common Spreadsheet File")
        fnparam.setValue(path)
        self.cache.update("common_spreadsheet_file", path)
        logger.info("common_spreadsheet_file set to {}".format(path))
        # print("common_spreadsheet_file set to {}".format(path))

    def get_parameter(self, param_path, parse_path=True, return_value=True):
        if parse_path:
            param_path = param_path.split(";")
        fnparam = self.parameters.param(*param_path)
        if return_value:
            return fnparam.value()
        else:
            return fnparam

    def set_parameter(self, param_path, value, parse_path=True):
        """
        Set value to parameter.
        :param param_path: Path to parameter can be separated by ";"
        :param value:
        :param parse_path: Turn on separation of path by ";"
        :return:
        """
        logger.debug(f"Set {param_path} to {value}")
        if parse_path:
            param_path = param_path.split(";")
        fnparam = self.parameters.param(*param_path)
        fnparam.setValue(value)

    def select_output_dir_gui(self):
        from PyQt5 import QtWidgets

        default_dir = self._prepare_default_output_dir()
        if op.exists(default_dir):
            start_dir = default_dir
        else:
            start_dir = op.dirname(default_dir)

        fn = QtWidgets.QFileDialog.getExistingDirectory(
            None,
            "Select Output Directory",
            directory=start_dir,
            # filter="NanoZoomer Digital Pathology Image(*.ndpi)"
        )
        # print (fn)
        self.set_output_dir(fn)

    def select_output_spreadsheet_gui(self):
        from PyQt5 import QtWidgets

        default_dir = self._prepare_default_output_dir()
        if op.exists(default_dir):
            start_dir = default_dir
        else:
            start_dir = op.dirname(default_dir)

        fn = QtWidgets.QFileDialog.getSaveFileName(
            None,
            "Select Common Spreadsheet File",
            directory=start_dir,
            filter="Excel File (*.xlsx)",
        )[0]
        # print (fn)
        self.set_common_spreadsheet_file(fn)

    def _prepare_default_output_common_spreadsheet_file(self):
        default_dir = io3d.datasets.join_path(get_root=True)
        # default_dir = op.expanduser("~/data")
        if not op.exists(default_dir):
            default_dir = op.expanduser("~")

        # timestamp = datetime.datetime.now().strftime("SA_%Y-%m-%d_%H:%M:%S")
        # timestamp = datetime.datetime.now().strftime("SA_%Y%m%d_%H%M%S")
        default_dir = op.join(default_dir, "SA_data.xlsx")
        return default_dir

    def _prepare_default_output_dir_prefix(self):
        default_dir = io3d.datasets.join_path(get_root=True)
        # default_dir = op.expanduser("~/data")
        if not op.exists(default_dir):
            default_dir = op.expanduser("~")
        return default_dir

    def _prepare_default_output_dir(self):

        # timestamp = datetime.datetime.now().strftime("SA_%Y-%m-%d_%H:%M:%S")
        timestamp = datetime.datetime.now().strftime("SA_%Y%m%d_%H%M%S")
        default_dir = op.join(self._default_output_dir_prefix, timestamp)
        return default_dir

    def parameters_to_dict(self):
        from . import dilipg

        return dilipg.params_and_values(self.parameters)

    def init_run(self):
        logger.debug("Init Run")
        fnparam = self.parameters.param("Input", "File Path")
        path = fnparam.value()
        # run_resc_int = self.parameters.param("Processing", "Intensity Normalization", "Run Intensity Normalization").value()
        self.anim = image.AnnotatedImage(path)
        self.intensity_rescale.set_anim_params(self.anim)
        fnparam = self.parameters.param("Output", "Directory Path")
        self.report.init_with_output_dir(fnparam.value())
        logger.debug(f"report output dir: {self.report.outputdir}")
        fn_spreadsheet = self.parameters.param("Output", "Common Spreadsheet File")
        self.report.additional_spreadsheet_fn = str(fn_spreadsheet.value())
        if self.get_parameter("Processing;SNI Prediction CNN") ==  True:
            self.lobule_quality_estimation_cnn.init()

    def set_annotation_color_selection(
        self, color: str, override_automatic_lobulus_selection=True
    ):
        if override_automatic_lobulus_selection:
            logger.debug("forced to use color selection")
            self.set_parameter("Input;Lobulus Selection Method", "Color")
        logger.debug(f"color={color}")
        pcolor = self.parameters.param("Input", "Annotation Color")
        color = color.upper()
        color_name = color.lower()
        color_name = color_name.capitalize()
        color_names = dict(zip(*pcolor.reverse[::-1]))
        if color_name in color_names:
            color = color_names[color_name]

        # rewrite name to code
        if color in pcolor.reverse[0]:
            # val = pcolor.reverse[0].index(color)
            # pcolor.setValue(val)
            logger.debug(f"setting color parameter to {color}")
            pcolor.setValue(color)
        else:
            raise ValueError("Color '{}' not found in allowed colors.".format(color))

    def train_scan_segmentation(
        self, fns: List[Union[str, Path]], clean_before_training=True
    ):
        """
        Train scan segmentation based on list of files with annotation.
        Output dir set before processing is ignored.
        :param fns: list of filenames
        :return:
        """
        #     mainapp = scaffan.algorithm.Scaffan()
        # if clf_fn is not None:
        #     mainapp.slide_segmentation.clf_fn = Path(clf_fn)
        # clf_fn = Path(mainapp.slide_segmentation.clf_fn)
        clf_fn = Path(self.slide_segmentation.clf_fn)
        logger.debug(f"train clf on path: {clf_fn}, exists: {clf_fn.exists()}")
        logger.debug(f"default output dir prefix={self._default_output_dir_prefix}")

        for i, fn in enumerate(fns):
            self.set_input_file(fn)
            self.set_output_dir()
            # There does not have to be set some color
            # mainapp.set_annotation_color_selection("#FF00FF")
            # mainapp.set_annotation_color_selection("#FF0000")
            #             mainapp.set_annotation_color_selection("#FFFF00")
            paramlist = [
                ("Input;Lobulus Selection Method", "None"),
                ("Processing;Skeleton Analysis", False),
                ("Processing;Texture Analysis", False),
                ("Processing;Lobulus Segmentation", False),
                ("Processing;Scan Segmentation;Run Prediction", False)
            ]
            prev_values = []
            for key, val in paramlist:
                prev_values.append(self.get_parameter(key))
                self.set_parameter(key, val)

            # self.set_parameter("Processing;Skeleton Analysis", False)
            # self.set_parameter("Processing;Texture Analysis", False)
            # self.set_parameter("Processing;Lobulus Segmentation", False)
            # self.set_parameter("Processing;Scan Segmentation;Run Prediction", False)
            if i == 0:
                if clean_before_training is not None:
                    self.set_parameter(
                        "Processing;Scan Segmentation;TFS General;Clean Before Training",
                        clean_before_training,
                    )
            else:
                self.set_parameter(
                    "Processing;Scan Segmentation;TFS General;Clean Before Training", False
                )
            self.set_parameter("Processing;Scan Segmentation;TFS General;Run Training", True)
            #             mainapp.set_parameter("Processing;Slide Segmentation;Lobulus Number", 0)
            # mainapp.start_gui(qapp=qapp)
            self.run_lobuluses()

            # reset values to prev values before function run
            for wanted, val in zip(paramlist, prev_values):
                key, wanted_val = wanted
                self.set_parameter(key, val)

    def _prepare_annoataion_ids_for_run_lobuluses(self, annotation_ids:Optional[list]=None):
        run_slide_segmentation = self.parameters.param(
            "Processing", "Scan Segmentation"
        ).value()
        pcolor = self.parameters.param("Input", "Annotation Color")
        color = pcolor.value()
        automatic_lobulus_selection = self.parameters.param(
            "Input", "Lobulus Selection Method"
        ).value()
        logger.debug(f"Lobulus Selection={automatic_lobulus_selection}")
        if annotation_ids is None:

            if automatic_lobulus_selection == "Color":
                annotation_ids = self.anim.get_annotations_by_color(
                    color,
                    raise_exception_if_not_found=self.raise_exception_if_color_not_found,
                )
            # elif automatic_lobulus_selection and not run_slide_segmentation:
            elif automatic_lobulus_selection == "Manual":
                # on training we dont want to do the prediction and we set
                annotation_ids = self.manual_select()
            elif automatic_lobulus_selection == "None":
                annotation_ids = []
            elif automatic_lobulus_selection == "Auto":
                if run_slide_segmentation:
                    self.slide_segmentation.add_biggest_to_annotations()
                    annotation_ids = self.slide_segmentation.ann_biggest_ids
                else:
                    logger.error("Cannot do automatic lobulus selection without whole scan segmentation")
            else:
                logger.error(f"Unknown Lobulus Selection Method: {automatic_lobulus_selection}")
        return annotation_ids, automatic_lobulus_selection

    def run_lobuluses(self, event=None, seeds_mm:Optional[list]=None):
        """
        :param event:
        :param seeds_mm: If not none, and Lobulus Selection Method is set to
        :return:
        """
        annotation_ids = None
        self.init_run()
        if seeds_mm:
            logger.debug(f"seeds_mm={seeds_mm}")
            _, annotation_ids = self.prepare_circle_annotations_from_seeds_mm(centers_mm=seeds_mm)

        self.report.level = self.parameters.param("Processing", "Report Level").value()

        show = self.parameters.param("Processing", "Show").value()
        self.report.set_show(show)
        self.report.set_save(True)
        run_slide_segmentation = self.parameters.param(
            "Processing", "Scan Segmentation"
        ).value()

        # if automatic_lobulus_selection == "use annotation_ids": ...

        t0 = time.time()
        if run_slide_segmentation:
            # fn_input = self.parameters.param("Input", "File Path").value()

            # TODO remove when whole scan will work fine
            # wsm = self.whole_scan_margin
            wsm = self.parameters.param("Processing", "Whole Scan Margin").value()
            self.slide_segmentation.init(self.anim.get_full_view(margin=wsm))
            self.slide_segmentation.run()
            # this can be used for automatic lobulus localization

        annotation_ids, automatic_lobulus_selection = self._prepare_annoataion_ids_for_run_lobuluses(
            annotation_ids
        )
        t1 = time.time()
        self.report.set_persistent_cols({"Lobules Selection Time [s]": t1 - t0})

        if annotation_ids is None:
            raise NoLobulusSelectionUsedError
        logger.debug("Annotation IDs: {}".format(annotation_ids))
        run_lob = self.parameters.param("Processing", "Lobulus Segmentation").value()
        if run_lob:
            for id in annotation_ids:
                self._run_lobulus(id)

        # in the case no lobulus has been measured the segmentation measurement is stored to table
        if not run_lob or (run_slide_segmentation and len(annotation_ids) == 0):
            self._add_general_information_to_actual_row()
            self.report.finish_actual_row()

        # self.report.df.to_excel(op.join(self.report.outputdir, "data.xlsx"))
        dumped = False
        while not dumped:
            try:
                # logger.debug(f"data frame {self.report.df}")
                logger.debug("Saving dataframe")
                self.report.dump()
                dumped = True
            except PermissionError as e:
                logger.error("Permission Error")
                if self.win is not None:
                    ret = QtGui.QMessageBox.warning(
                        self.win,
                        "XLSX file opened in external application",
                        "Close opened spreadsheet files before continue",
                    )

                else:
                    raise e
        saved_params = self.parameters.saveState()
        io3d.misc.obj_to_file(
            saved_params, str(Path(self.report.outputdir) / "parameters.yaml")
        )
        try:
            with open(
                str(Path(self.report.outputdir) / "parameters.json"), "w"
            ) as outfile:
                json.dump(saved_params, outfile)
        except:
            import traceback

            logger.debug("saved_params: " + str(saved_params))
            logger.warning(f"Problem with dump file to json: {traceback.format_exc()}")
        from . import os_interaction

        open_dir = self.parameters.param("Processing", "Open output dir").value()
        if open_dir:
            os_interaction.open_path(self.report.outputdir)
        logger.debug("finished")

        # print("ann ids", annotation_ids)

    def _add_general_information_to_actual_row(self, update_dict:dict=None):
        inpath = Path(self.parameters.param("Input", "File Path").value())
        fn = inpath.parts[-1]
        fn_out = self.parameters.param("Output", "Directory Path").value()
        data_to_add = {
            "File Name": str(fn),
            "File Path": str(inpath),
            "Annotation Color": self.parameters.param(
                "Input", "Annotation Color"
            ).value(),
            "Datetime": datetime.datetime.now().isoformat(" ", "seconds"),
            "platform.system": platform.uname().system,
            "platform.node": platform.uname().node,
            "platform.processor": platform.uname().processor,
            "Scaffan Version": scaffan.__version__,
            "Output Directory Path": str(fn_out),
        }
        if update_dict:
            data_to_add.update(update_dict)
        self.report.add_cols_to_actual_row(
            data_to_add
        )
        self.report.add_cols_to_actual_row(self.parameters_to_dict())

    def _run_lobulus(self, annotation_id):
        show = self.parameters.param("Processing", "Show").value()
        t0 = time.time()
        inpath = Path(self.parameters.param("Input", "File Path").value())
        fn = inpath.parts[-1]
        self.report.add_cols_to_actual_row(
            {
                # "File Name": str(fn),
                "Annotation ID": annotation_id,
            }
        )
        logger.info(f"Processing file: {fn} with Annotation ID: {annotation_id}")

        self.lobulus_processing.set_annotated_image_and_id(self.anim, annotation_id)
        self.lobulus_processing.run(show=show)
        logger.trace(
            f"type lobulus_processing.lobulus_mask: {type(self.lobulus_processing.lobulus_mask)}"
        )
        self.skeleton_analysis.set_lobulus(lobulus=self.lobulus_processing)
        logger.debug("set lobulus done")
        # run_slide_segmentation = self.parameters.param("Processing", "Texture Analysis").value()
        run_skeleton_analysis = self.parameters.param(
            "Processing", "Skeleton Analysis"
        ).value()
        run_texture_analysis = self.parameters.param(
            "Processing", "Texture Analysis"
        ).value()
        run_lobule_quality_estimation = self.parameters.param(
            "Processing", "Quality Estimation CNN"
        ).value()
        logger.debug("before skeleton analysis")
        if run_skeleton_analysis:
            self.skeleton_analysis.skeleton_analysis(show=show)
        if run_texture_analysis:
            # self.glcm_textures.report = self.report
            self.glcm_textures.set_input_data(
                view=self.lobulus_processing.view,
                annotation_id=annotation_id,
                lobulus_segmentation=self.lobulus_processing.lobulus_mask,
            )
            self.glcm_textures.run()
        if run_lobule_quality_estimation:
            self.lobule_quality_estimation_cnn.set_input_data(
                view=self.lobulus_processing.view,
                annotation_id=annotation_id,
                lobulus_segmentation=self.lobulus_processing.lobulus_mask,
            )
            self.lobule_quality_estimation_cnn.run()
        logger.trace("after texture analysis")
        t1 = time.time()
        ann_center = self.anim.get_annotation_center_mm(annotation_id)
        # inpath = Path(self.parameters.param("Input", "File Path").value())
        # fn = inpath.parts[-1]
        self._add_general_information_to_actual_row()
        self.report.add_cols_to_actual_row(
            {
                # "File Path": str(inpath),
                # "Annotation Color": self.parameters.param("Input", "Annotation Color").value(),
                "Annotation Center X [mm]": ann_center[0],
                "Annotation Center Y [mm]": ann_center[1],
                "Processing Time [s]": t1 - t0,
                # "Datetime": datetime.datetime.now().isoformat(' ', 'seconds'),
                # "platform.system": platform.uname().system,
                # "platform.node": platform.uname().node,
                # "platform.processor": platform.uname().processor,
                # "Scaffan Version": scaffan.__version__,
            }
        )
        # evaluation
        self.evaluation.set_input_data(
            self.anim, annotation_id, self.lobulus_processing
        )
        self.evaluation.run()
        # Copy all parameters to table
        self.report.finish_actual_row()

    def get_preview_view(self):
        full_view = self.anim.get_view(
            location=[0, 0], level=0, size_on_level=self.anim.get_slide_size()[::-1]
        )
        pxsz_mm = float(self.get_parameter("Processing;Preview Pixelsize")) * 1000
        view_corner = full_view.to_pixelsize(pixelsize_mm=[pxsz_mm, pxsz_mm])
        return view_corner

    def get_preview(self):
        view_corner = self.get_preview_view()
        # logger.debug(
        #     f"Manual selection1, view.loc={full_view.region_location}, view.size={full_view.region_size_on_level}, pxsz={full_view.region_pixelsize}"
        # )
        logger.debug(
            f"Manual selection2, view.loc={view_corner.region_location}, view.size={view_corner.region_size_on_level}, pxsz={view_corner.region_pixelsize}"
        )
        img = view_corner.get_region_image(as_gray=False)
        return view_corner, img

    def prepare_circle_annotations_from_seeds_mm(self, centers_mm):
        pxsz_mm = float(self.get_parameter("Processing;Preview Pixelsize")) * 1000
        centers_mm = list(np.asarray(centers_mm) / pxsz_mm)

        view_corner = self.get_preview_view()
        return self.prepare_circle_annotations_from_points_px_in_preview(view_corner, centers_mm)

    def prepare_circle_annotations_from_points_px_in_preview(self, view_corner, centers_px_view):

        x_px_view, y_px_view = zip(*list(centers_px_view))
        logger.debug(x_px_view)
        pts_glob_px = view_corner.coords_view_px_to_glob_px(np.asarray(x_px_view), np.asarray(y_px_view))
        centers_px = list(zip(*pts_glob_px))
        logger.debug(f"Manual selection5, centers_px_global={centers_px}")
        r_mm = (
                float(self.get_parameter("Processing;Scan Segmentation;Annotation Radius"))
                * 1000
        )

        ann_ids, _ = scaffan.slide_segmentation.add_circle_annotation(
            view_corner, centers_px, annotations=self.anim.annotations, r_mm=r_mm
        )
        view_corner.set_annotations(self.anim.annotations)
        view_corner.adjust_annotation_to_image_view()
        return view_corner, ann_ids

    def manual_select(self):
        logger.debug("Manual selection")
        # full_view = self.anim.get_full_view()
        view_corner, img = self.get_preview()
        # full_view = self.anim.get_view(
        #     location=[0, 0], level=0, size_on_level=self.anim.get_slide_size()[::-1]
        # )
        # pxsz_mm = float(self.get_parameter("Processing;Preview Pixelsize")) * 1000
        # view_corner = full_view.to_pixelsize(pixelsize_mm=[pxsz_mm, pxsz_mm])
        # logger.debug(
        #     f"Manual selection1, view.loc={full_view.region_location}, view.size={full_view.region_size_on_level}, pxsz={full_view.region_pixelsize}"
        # )
        # logger.debug(
        #     f"Manual selection2, view.loc={view_corner.region_location}, view.size={view_corner.region_size_on_level}, pxsz={view_corner.region_pixelsize}"
        # )
        # img = view_corner.get_region_image(as_gray=False)
        plt.ioff()
        fig = plt.figure(figsize=(12, 8))
        fig.canvas.set_window_title(
            "Select lobules. Left/Middle/Right Mouse Button: add/quit/remove"
        )
        logger.debug(
            f"Manual selection2 backend={matplotlib.get_backend()}, ion={matplotlib.is_interactive()}, img.shape={img.shape}, img.max={np.max(img)}"
        )
        plt.imshow(img)
        # plt.ginput(1)
        # plt.axis("image")
        # fig.ax
        # fig.axes[0].imshow(img)
        # logger.debug("Manual selection4")
        points_px = np.asarray(plt.ginput(-1))
        plt.close(fig)
        logger.debug(f"Manual selection5, centers_px={points_px}")

        # points_px = np.asarray(centers_px)
        from scaffan.image import get_offset_px, get_pixelsize

        # offset_px = get_offset_px(self.anim)
        # pixelsize, pixelunit = get_pixelsize(self.anim, requested_unit="mm")
        x_px_view = points_px[:, 0].flatten()
        y_px_view = points_px[:, 1].flatten()
        # x_px = (x_px_view*view_corner.zoom[0] + offset_px[0])
        # y_px = (y_px_view*view_corner.zoom[1] + offset_px[1])
        # pts_glob_px = [x_px, y_px]
        xy_px_view = list(zip(x_px_view, y_px_view))

        pxsz_mm = float(self.get_parameter("Processing;Preview Pixelsize")) * 1000
        seeds_mm = list(np.asarray(xy_px_view) / pxsz_mm)
        logger.debug(f"seeds_mm={seeds_mm}")
        # centers_px = list(zip(*pts_glob_px))
        view_corner, ann_ids = self.prepare_circle_annotations_from_points_px_in_preview(
            view_corner, centers_px_view=xy_px_view)
        # logger.debug(f"annotations={self.anim.annotations}")

        logger.debug("annotation selected")
        fig = plt.figure()
        plt.imshow(img)
        view_corner.plot_annotations(None)
        centers_px = plt.ginput(1)
        plt.draw()
        plt.show()
        plt.draw()
        plt.close(fig)
        return ann_ids

    def set_persistent_cols(self, dct: dict):
        """
        Set data which will be appended to all rows.
        :param dct: dictionary with column name and value
        :return:
        """
        self.report.set_persistent_cols(dct)

    def _get_file_info(self):
        fnparam = Path(self.parameters.param("Input", "File Path").value())
        if fnparam.exists() and fnparam.is_file():
            anim = scaffan.image.AnnotatedImage(str(fnparam))
            self.parameters.param("Input", "Data Info").setValue(anim.get_file_info())
            # self.parameters.param("Input", "Select").setValue(anim.get_file_info())

    def start_gui(self, skip_exec=False, qapp=None):

        from PyQt5 import QtWidgets
        import scaffan.qtexceptionhook

        # import QApplication, QFileDialog
        if not skip_exec and qapp == None:
            qapp = QtWidgets.QApplication(sys.argv)

        self.parameters.param("Input", "Select").sigActivated.connect(
            self.select_file_gui
        )
        self.parameters.param("Output", "Select").sigActivated.connect(
            self.select_output_dir_gui
        )
        self.parameters.param(
            "Output", "Select Common Spreadsheet File"
        ).sigActivated.connect(self.select_output_spreadsheet_gui)
        self.parameters.param("Run").sigActivated.connect(self.run_lobuluses)

        self.parameters.param("Processing", "Open output dir").setValue(True)
        t = ParameterTree()
        t.setParameters(self.parameters, showTop=False)
        t.setWindowTitle("pyqtgraph example: Parameter Tree")
        t.show()

        # print("run scaffan")
        win = QtGui.QWidget()
        win.setWindowTitle("ScaffAn {}".format(scaffan.__version__))
        logo_fn = op.join(op.dirname(__file__), "scaffan_icon256.png")
        app_icon = QtGui.QIcon()
        # app_icon.addFile(logo_fn, QtCore.QSize(16, 16))
        app_icon.addFile(logo_fn)
        win.setWindowIcon(app_icon)
        # qapp.setWindowIcon(app_icon)
        layout = QtGui.QGridLayout()
        win.setLayout(layout)
        pic = QtGui.QLabel()
        pic.setPixmap(QtGui.QPixmap(logo_fn).scaled(100, 100))
        pic.show()
        # layout.addWidget(QtGui.QLabel("These are two views of the same data. They should always display the same values."), 0,  0, 1, 2)
        layout.addWidget(pic, 1, 0, 1, 1)
        layout.addWidget(t, 2, 0, 1, 1)
        # layout.addWidget(t2, 1, 1, 1, 1)

        win.show()
        win.resize(800, 800)
        self.win = win
        # win.
        if not skip_exec:

            qapp.exec_()


class NoLobulusSelectionUsedError(Exception):
    pass



#! /usr/bin/python
# -*- coding: utf-8 -*-


from loguru import logger

import zipfile
import os
import os.path as op
import tempfile
import sys

if sys.version_info < (3, 0):
    import urllib as urllibr
else:
    import urllib.request as urllibr


def download_and_unzip(url, outdir):
    # try:
    #     import pywget
    # except:
    #     import wget as pywget

    if outdir is None:

        outdir = tempfile.gettempdir()
        # print("temp directory ", outdir)
        outdir = tempfile.mkdtemp()
    # there is problem with wget. It uses its ouwn tempfile in current dir. It is not sure that there will
    # be requred permisson for write

    if not op.exists(outdir):
        os.makedirs(outdir)
    filename = op.join(outdir, "openslides.zip")
    urllibr.urlretrieve(url, filename)
    # filename = pywget.download(url, out=filename)

    zf = zipfile.ZipFile(filename)
    zf.extractall(outdir)
    zf.close()
    return outdir


def libfix():

    # glob.glob(op.expanduser("~/Downloads/"))
    if sys.platform.startswith("win"):
        # print("Trying to download .dll libraries")
        libfix_windows()
    if sys.platform.startswith("linux"):
        # print("Trying to download .so libraries")
        pass
        # libfix_linux_conda()


# def get_conda_dir():
#     """
#     Get conda base dir.
#     I would remove envs subdirs from output dir.
#
#     :return:
#     """
#     if sys.version_info.major == 2:
#         conda_dir = get_conda_dir_old()
#     else:
#         conda_dir = sys.exec_prefix
#     idx = conda_dir.find("envs")
#     if idx > 0:
#         conda_dir = conda_dir[:idx]
#     return conda_dir


def libfix_windows(url=None):
    if url is None:
        if sys.maxsize > 2 ** 32:
            url = "https://github.com/openslide/openslide-winbuild/releases/download/v20171122/openslide-win64-20171122.zip"
        else:
            url = "https://github.com/openslide/openslide-winbuild/releases/download/v20171122/openslide-win32-20171122.zip"
    outdir = download_and_unzip(url, op.expanduser("~/Downloads/"))

    # dest_dir = get_conda_dir()

    # for file in glob.glob(r'ITK+Skelet3D_dll/*.dll'):
    #     shutil.copy(file, dest_dir)
    #     print("copy %s ---> %s" % (file, dest_dir))
    #
    # try:
    #     shutil.rmtree(outdir)
    # except:
    #     import traceback
    #     traceback.print_exc()



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Modul is used for GUI of Lisa
"""

from loguru import logger
import os.path as op
import sys

from scaffan import annotation

# print("ann to json")

if __name__ == "__main__":

    # print(sys.argv)
    pth = op.expanduser(sys.argv[1])
    # print(pth)
    annotation.ndpa_to_json(pth)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © %YEAR%  <>
#
# Distributed under terms of the %LICENSE% license.

"""
"""

from loguru import logger
import argparse


from PyQt5 import QtGui
import sys

import collections
import numpy as np

from pyqtgraph.parametertree import ParameterTree
import pyqtgraph.parametertree.parameterTypes as pTypes


class BatchFileProcessingParameter(pTypes.GroupParameter):
    def __init__(self, **opts):
        opts["type"] = "group"
        opts["addText"] = "Add"
        {"name": "Save State", "type": "action"},
        # opts['addList'] = ['str', 'float', 'int']
        pTypes.GroupParameter.__init__(self, **opts)

    def addNew(self):
        # val = {
        #     'str': '',
        #     'float': 0.0,
        #     'int': 0
        # }[typ]
        fname = QtGui.QFileDialog.getOpenFileName(None, "Open file", "")
        self.add_filename(fname=fname)

    def add_filename(self, fname):
        self.addChild(
            dict(
                name="%i" % ((len(self.childs))),
                type="str",
                value=str(fname),
                removable=True,
                renamable=True,
            )
        )



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Process lobulus analysis.
"""

from loguru import logger
import skimage.filters
from skimage.morphology import skeletonize
import skimage.measure
import skimage.io
import copy
import scipy.signal
import scipy.ndimage
import os.path as op

import numpy as np
import warnings
import morphsnakes as ms
from matplotlib import pyplot as plt
from scaffan import image as scim
from exsu.report import Report
from pyqtgraph.parametertree import Parameter
import imma.image
import scaffan.texture

_cite = (
    ""
    + "[1]: A Morphological Approach to Curvature-based Evolution of Curves and Surfaces, Pablo Márquez-Neila, Luis Baumela and Luis Álvarez. In IEEE Transactions on Pattern Analysis and Machine Intelligence (PAMI), 2014, DOI 10.1109/TPAMI.2013.106"
    + "[2]: Morphological Snakes. Luis Álvarez, Luis Baumela, Pablo Márquez-Neila. In Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition 2010 (CVPR10)."
)


class Lobulus:
    def __init__(
        self,
        pname="Lobulus Segmentation",
        ptype="group",
        pvalue=True,
        ptip="Select the area for sinusoidal area texture analysis after lobuli selection",
        pexpanded=False,
        report: Report = None,
    ):
        # TODO the segmentation resolution was probably different.
        #  For segmentation was used different level than 2. Probably 3 or 4
        #  The level 2 has been used in detail view

        self._inner_texture = scaffan.texture.GLCMTextureMeasurement(
            "central_vein",
            texture_label="central_vein",
            add_cols_to_report=False,
            report=report,
        )
        params = [
            # {
            #     "name": "Tile Size",
            #     "type": "int",
            #     "value" : 128
            # },
            {
                "name": "Working Resolution",
                "type": "float",
                # "value": 0.000001,
                # "value": 0.000001,  # this is typical resolution on level 2-3
                # "value": 0.000002,  # this is typical resolution on level 3-4
                # "value": 0.0000015,  # this is typical resolution on level 3-4
                # "value": 0.000000227 # level 0
                # "value": 0.000000454 # level 1
                # "value": 0.000000908 # level 2
                # "value": 0.00000091, # this is typical resolution on level 2
                "value": 0.00000182,  # this is typical resolution on level 3 # probably prev stable version
                # "value": 0.00000364,  # this is typical resolution on level 4
                # "value": 0.00000728,  # this is typical resolution on level 5
                # "value": 0.00001456,  # this is typical resolution on level 6
                "suffix": "m",
                "siPrefix": True,
            },
            {
                "name": "Annotation Margin",
                "type": "float",
                "value": 180,
                "suffix": "%",
                "siPrefix": False,
                "tip": "Margin added to the input annotation where is expected whole lobulus.",
            },
            {
                "name": "Border Segmentation",
                "type": "group",
                "tip": "MorphACWE algorithm parameters. " + _cite,
                "children": [
                    {
                        "name": "Smoothing",
                        "type": "int",
                        "value": 2,
                        "suffix": "px",
                        "siPrefix": False,
                        "tip": "MorphACWE algorithm parameter: The number of repetitions of the smoothing step (the curv operator) in each iteration. In other terms, this is the strength of the smoothing. This is the parameter µ.",
                    },
                    {
                        "name": "Lambda1",
                        "type": "float",
                        "value": 1.0,
                        "tip": "MorphACWE algorithm parameter: Relative importance of the inside pixels (lambda1) against the outside pixels (lambda2).",
                        # "suffix": "px",
                        # "siPrefix": False
                    },
                    {
                        "name": "Lambda2",
                        "type": "float",
                        "value": 2.0,
                        "tip": "MorphACWE algorithm parameter: Relative importance of the inside pixels (lambda1) against the outside pixels (lambda2).",
                        # "suffix": "px",
                        # "siPrefix": False
                    },
                    {
                        "name": "Iterations",
                        "type": "int",
                        # "value": 180
                        "value": 400,
                    },
                ],
            },
            {
                "name": "Central Vein Segmentation",
                "type": "group",
                "tip": "MorphGAC algorithm parameters. " + _cite,
                "children": [
                    {
                        "name": "Smoothing",
                        "type": "int",
                        "value": 2,
                        "suffix": "px",
                        "siPrefix": False,
                        "tip": "MorphGAC algorithm parameter: The number of repetitions of the smoothing step in each iteration. This is the parameter µ.",
                    },
                    {
                        "name": "Threshold",
                        "type": "float",
                        # "value": 0.28, # prev stable version
                        # "value": 0.35,
                        # "value": 0.25,
                        "value": 0.22,
                        # "value": 0.25,
                        "tip": "MorphGAC algorithm parameter: The threshold that determines which areas are affected by the morphological balloon. This is the parameter θ.",
                        # "suffix": "px",
                        # "siPrefix": False
                    },
                    {
                        "name": "Ballon",
                        "type": "float",
                        "value": -1,
                        "tip": "MorphGAC algorithm parameter: The strength of the morphological balloon. This is the parameter ν.",
                        # "suffix": "px",
                        # "siPrefix": False
                    },
                    {
                        "name": "Iterations",
                        "type": "int",
                        # "value": 400
                        "value": 150,
                    },
                    self._inner_texture.parameters,
                ],
            },
            {
                "name": "Manual Segmentation",
                "type": "bool",
                "value": False,
                "tip": "Use manual inner and outer segmentation with black color instead of computed segmentation. Computation is skipped.",
            },
        ]

        # self.parameters = Parameter.create(
        #     name="Lobulus Segmentation", type="group",
        #     children=params, expanded=False)
        self.parameters = Parameter.create(
            name=pname,
            type=ptype,
            value=pvalue,
            tip=ptip,
            children=params,
            expanded=pexpanded,
        )
        self.report: Report = report

    def set_annotated_image_and_id(
        self, anim: scim.AnnotatedImage, annotation_id, level=None
    ):
        """

        :param anim:
        :param annotation_id:
        :param level: Is used just for short time in funcion get_views()
        :return:
        """
        self.anim = anim
        self.level = level
        self.annotation_id = annotation_id
        self._init_by_annotation_id(annotation_id)

    # def set_report(self, report: Report):
    #     self.report: Report = report

    def _init_by_annotation_id(self, annotation_id):
        if not np.isscalar(annotation_id):
            raise ValueError(
                "Annotation ID should be scalar int value for lobulus processing."
            )

        pixelsize_mm = [
            (self.parameters.param("Working Resolution").value() * 1000)
        ] * 2
        annotation_margin = self.parameters.param("Annotation Margin").value() * 0.01
        self.view = self.anim.get_views(
            annotation_ids=[annotation_id],
            level=self.level,
            margin=annotation_margin,
            pixelsize_mm=pixelsize_mm,
        )[0]
        # right_shape = imma.image.calculate_new_shape(
        #     self.view.region_size_on_level,
        #     self.view.get_pixelsize_on_level(self.view.region_level)[0],
        #     self.view.region_pixelsize
        # )
        self.image = self.view.get_region_image(as_gray=True)
        self.annotation_mask = self.view.get_annotation_raster(
            annotation_id=annotation_id
        )
        self._im_gradient_border_frangi = None
        # self.anim.titles
        # self.anim.details
        pass

    def find_border(self, show=True):
        outer_ids = self.anim.select_outer_annotations(
            self.annotation_id, color="#000000", raise_exception_if_not_found=False
        )
        if len(outer_ids) > 1:
            logger.warning(
                f"More than one outer annotation find to annotation with ID: {self.annotation_id}"
            )
        elif len(outer_ids) > 0:
            outer_id = outer_ids[0]
            seg_true = self.view.get_annotation_raster(annotation_id=outer_id) > 0
            use_manual = self.parameters.param(
                # "Processing", "Lobulus Segmentation",
                "Manual Segmentation"
            ).value()
            # import pdb
            # pdb.set_trace()
            if use_manual:

                self.border_mask = seg_true.astype(np.uint8)
                logger.debug("Manual segmentation ")
                return
        self._im_gradient_border_frangi = skimage.filters.frangi(self.image)
        logger.debug("Image size {}".format(self.image.shape))
        circle = self.annotation_mask
        if self.report is not None:
            self.report.imsave_as_fig(
                f"gradient_outer_{self.annotation_id}.png",
                self._im_gradient_border_frangi,
                level=30,
            )
        param_acwe_smoothing = self.parameters.param(
            "Border Segmentation", "Smoothing"
        ).value()
        param_acwe_lambda1 = self.parameters.param(
            "Border Segmentation", "Lambda1"
        ).value()
        param_acwe_lambda2 = self.parameters.param(
            "Border Segmentation", "Lambda2"
        ).value()
        param_acwe_iterations = self.parameters.param(
            "Border Segmentation", "Iterations"
        ).value()
        mgac = ms.MorphACWE(
            self._im_gradient_border_frangi,
            smoothing=param_acwe_smoothing,
            lambda1=param_acwe_lambda1,
            lambda2=param_acwe_lambda2,
        )
        mgac.levelset = circle.copy()
        mgac.run(iterations=param_acwe_iterations)
        outer = mgac.levelset.copy()
        self.border_mask = outer.astype(np.uint8)

    def find_central_vein(self, show=True):
        inner_ids = self.anim.select_inner_annotations(
            self.annotation_id, color="#000000"
        )
        if len(inner_ids) > 1:
            logger.warning(
                "More than one inner annotation find to annotation with ID %i. Combination will be used.",
                self.annotation_id,
            )
        if len(inner_ids) > 0:
            # get first
            inner_id = inner_ids[0]
            seg_true = self.view.get_annotation_raster(annotation_id=inner_id) > 0
            for inner_id in inner_ids:
                seg_true += self.view.get_annotation_raster(annotation_id=inner_id) > 0
            seg_true = (seg_true > 0).astype(np.uint8)

            use_manual = self.parameters.param(
                # "Processing", "Lobulus Segmentation",
                "Manual Segmentation"
            ).value()
            if use_manual:
                # self.central_vein_mask = seg_true.astype(np.float64)
                self.central_vein_mask = seg_true.astype(np.uint8)
                return
        use_texture_features = True
        pixelsize_mm = [
            (self.parameters.param("Working Resolution").value() * 1000)
        ] * 2
        central_vein_view = self.view.anim.get_views(
            [self.annotation_id], pixelsize_mm=pixelsize_mm, margin=0.1
        )[0]

        self._inner_texture.set_input_data(
            view=central_vein_view,
            annotation_id=self.annotation_id,
            lobulus_segmentation=None,
        )
        # self._inner_texture.set_report(self.report)
        self._inner_texture.add_cols_to_report = False
        self._inner_texture.run()
        tfeatures = copy.copy(self._inner_texture.measured_features)
        tfeatures[:, :, 2] = 1 - tfeatures[:, :, 2]
        tfeatures0 = np.mean(tfeatures, 2)
        if self.report is not None:
            self.report.imsave_as_fig(
                "gradient_texture_color_{}.png".format(self.annotation_id),
                tfeatures,
                level=45,
            )
            self.report.imsave_as_fig(
                "gradient_texture_mean_{}.png".format(self.annotation_id),
                tfeatures0,
                level=40,
            )

        sl = self.view.get_slices_for_insert_image_from_view(central_vein_view)

        image = central_vein_view.get_region_image(as_gray=True)
        if self._im_gradient_border_frangi is None:
            self._im_gradient_border_frangi = skimage.filters.frangi(self.image)
        im_gradient_border_frangi = self._im_gradient_border_frangi[
            sl
        ]  # skimage.filters.frangi(image)
        # gborders us internally: scipy.ndimage.filters.gaussian_gradient_magnitude
        im_gradient_base_inner = ms.gborders(image, alpha=1000, sigma=2)
        im_gradient_inner = im_gradient_base_inner - (im_gradient_border_frangi * 10000)
        if use_texture_features:
            im_gradient_inner *= 1 - imma.image.resize_to_shape(
                tfeatures0, im_gradient_inner.shape
            )
        if self.report is not None:
            self.report.imsave_as_fig(
                "gradient_inner_frangi_{}.png".format(self.annotation_id),
                im_gradient_border_frangi,
                level=35,
            )
            self.report.imsave_as_fig(
                "gradient_base_inner_{}.png".format(self.annotation_id),
                im_gradient_base_inner,
                level=35,
            )
            self.report.imsave_as_fig(
                "gradient_inner_{}.png".format(self.annotation_id),
                im_gradient_inner,
                level=35,
            )
        # circle = circle_level_set(imgr.shape, size2, 75, scalerow=0.75)
        circle = self.annotation_mask[sl]
        # plt.figure()
        # plt.imshow(im_gradient0)
        # plt.colorbar()
        # plt.contour(circle)
        # plt.show()
        # mgac = ms.MorphGAC(im_gradient, smoothing=2, threshold=0.3, balloon=-1)
        # mgac.levelset = circle.copy()
        # mgac.run(iterations=100)
        # inner = mgac.levelset.copy()

        param_gac_smoothing = self.parameters.param(
            "Central Vein Segmentation", "Smoothing"
        ).value()
        param_gac_threshold = self.parameters.param(
            "Central Vein Segmentation", "Threshold"
        ).value()
        param_gac_baloon = self.parameters.param(
            "Central Vein Segmentation", "Ballon"
        ).value()
        param_gac_iterations = self.parameters.param(
            "Central Vein Segmentation", "Iterations"
        ).value()

        # central vein
        mgac = ms.MorphGAC(
            im_gradient_inner,
            smoothing=param_gac_smoothing,
            threshold=param_gac_threshold,
            balloon=param_gac_baloon,
        )
        # mgac = ms.MorphGAC(im_gradient, smoothing=2, threshold=0.28, balloon=-1.0)
        # mgac = ms.MorphACWE(im_gradient0, smoothing=2, lambda1=.1, lambda2=.05)
        mgac.levelset = circle.copy()
        mgac.run(iterations=param_gac_iterations)
        inner = mgac.levelset.copy()

        if self.report is not None:
            fig = plt.figure(figsize=(12, 10))
            plt.imshow(im_gradient_inner)
            plt.colorbar()
            plt.contour(circle + inner)
            self.report.savefig_and_show(
                "lobulus_gradient_inner_{}.png".format(self.annotation_id), fig=fig
            )

        cvmask = np.zeros_like(self.border_mask)
        cvmask = self.view.insert_image_from_view(central_vein_view, cvmask, inner)
        # cvmask[sl] = inner
        self.central_vein_mask = cvmask

    def run(self, show=True):

        self.find_border(show)
        self.find_central_vein(show)
        # inner_lobulus_margin_mm = 0.02

        # mgac = ms.MorphGAC(im_gradient, smoothing=2, threshold=0.2, balloon=+1)
        # mgac = ms.MorphACWE(im_gradient0, smoothing=2, lambda1=0.5, lambda2=1.0)

        # circle = circle_level_set(imgr.shape, (200, 200), 75, scalerow=0.75)

        # plt.figure()
        # plt.imshow(im_gradient)
        # plt.colorbar()
        # plt.contour(circle + inner + outer)
        fig = plt.figure(figsize=(12, 10))
        plt.imshow(self.image, cmap="gray")
        plt.colorbar()
        plt.contour(self.annotation_mask + self.central_vein_mask + self.border_mask)
        self.view.add_ticks()

        datarow = {}
        datarow["Annotation ID"] = self.annotation_id
        numeric_id = self.anim.get_annotation_id(self.annotation_id)

        # self.anim.annotations.
        datarow["Annotation Title"] = self.anim.annotations[numeric_id][
            "title"
        ]  # [self.annotation_id]
        datarow["Annotation Details"] = self.anim.annotations[numeric_id][
            "details"
        ]  # [self.annotation_id]
        # datarow["Annotation Details"] = self.anim.details[self.annotation_id]
        if self.report is not None:
            self.report.savefig_and_show(
                "lobulus_{}.png".format(self.annotation_id), fig=fig, level=80
            )

        self.lobulus_mask = (self.central_vein_mask + self.border_mask) == 1
        if self.report:
            self.report.imsave_as_fig(
                f"lobulus_mask_{self.annotation_id}.png",
                self.lobulus_mask.astype(np.uint8),
                level=70,
                # level_skimage=20, npz_level=30
            )
        area_px = np.sum(self.lobulus_mask)
        datarow["Area"] = area_px * np.prod(self.view.region_pixelsize)
        # rprops = skimage.measure.regionprops(self.lobulus_mask)
        # logger.debug(f"len rprops: {len(rprops)}")
        perimeter_px = skimage.measure.perimeter(self.lobulus_mask, neighbourhood=8)
        datarow["Lobulus Perimeter"] = perimeter_px * self.view.region_pixelsize[0]
        datarow["Lobulus Boundary Compactness"] = (
            4 * np.pi * area_px / perimeter_px ** 2
        )
        datarow["Lobulus Equivalent Diameter"] = datarow["Lobulus Perimeter"] / np.pi
        datarow["Lobulus Equivalent Surface"] = (
            np.pi * datarow["Lobulus Equivalent Diameter"] ** 2 / 4.0
        )

        datarow["Central vein area"] = np.sum(self.central_vein_mask > 0) * np.prod(
            self.view.region_pixelsize
        )
        datarow["Area unit"] = self.view.region_pixelunit
        if self.report:
            self.report.add_cols_to_actual_row(datarow)
        # self.skeleton_analysis(show=show)

    def imsave(self, base_fn, arr, k=50):
        base_fn = base_fn.format(self.annotation_id)
        self.report.imsave(base_fn, arr, k)



# /usr/bin/env python
# -*- coding: utf-8 -*-
"""
Modul is used for GUI of Lisa
"""

from loguru import logger
import sys
import click
from pathlib import Path
import ast

# print("start")
# from . import image

# print("start 5")
# print("start 6")

from scaffan import algorithm
from . import app_tools

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


# print("Running __main__.py")
# @batch_detect.command(context_settings=CONTEXT_SETTINGS)
# @click.argument("image_stack_dir", type=click.Path(exists=True))
# @click.argument("working_dir", type=click.Path())
# @click.option("--create-icon", is_flag=True,
#               help="Create desktop icon"
#               )
@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.option(
    "--log-level",
    "-ll",
    # type=,
    help="Set logging level",
    default="INFO",
)
@click.option(
    "--log-file",
    "-lf",
    # type=,
    help="Set logging file",
    default=None,
)
@click.pass_context
def run(ctx, log_level, log_file, *args, **kwargs):
    if log_level is not None:
        try:
            log_level = int(log_level)
        except ValueError as e:
            log_level = log_level.upper()
        # logger se odebírá při debugování v pycharm
        # if log_level == "DEBUG":
        #     pass
        # else:
        if log_file:
            Path(log_file).parent.mkdir(parents=True, exist_ok=True)
            sink = log_file
        else:
            sink = sys.stderr
        logger.remove()
        i = logger.add(sink, level=log_level, colorize=True)
    if ctx.invoked_subcommand is None:
        # click.echo("I am about to invoke GUI")
        ctx.invoke(gui, *args, **kwargs)
    else:
        logger.debug(f"I am about to invoke {ctx.invoked_subcommand}")
        # next command is useless. It is invoked automatically
        # ctx.invoke(ctx.invoked_subcommand, *args, **kwargs)
        # pass


@run.command(context_settings=CONTEXT_SETTINGS, help="Set persistent values")
@click.option(
    "--common-spreadsheet-file",
    help="Set path for common spreadsheet file.",
    type=click.Path(),
)
def set(common_spreadsheet_file=None):
    mainapp = algorithm.Scaffan()
    if common_spreadsheet_file is not None:
        mainapp.set_common_spreadsheet_file(path=common_spreadsheet_file)
        logger.info(f"Common spreadsheet file path is : {common_spreadsheet_file}")
        print(f"Common spreadsheet file path is : {common_spreadsheet_file}")


# def print_params(params):
#     algorithm.Scaffan().parameters.
#     params.


@run.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--params",
    "-p",
    multiple=True,
    default="",
    nargs=2,
    help='Set parameter. First argument is path to parameter separated by ";". Second is the value.'
    "python -m scaffan gui -p Processing,Show True",
)
# @click.option(
#     "--log-level",
#     "-ll",
#     # type=,
#     help="Set logging level",
#     default=None,
# )
@click.option("--print-params", "-pp", is_flag=True, help="Print parameters")
def gui(params, print_params):
    mainapp = algorithm.Scaffan()
    if print_params:
        import pprint

        pprint.pprint(mainapp.parameters_to_dict())
        exit()
    for param in params:
        mainapp.set_parameter(param[0], value=ast.literal_eval(param[1]))
        # mainapp.parameters.param(*param[0].split(";")).setValue(ast.literal_eval(param[1]))
    mainapp.start_gui()


@run.command(
    context_settings=CONTEXT_SETTINGS, help="Create an icon on Windows platform"
)
def install():
    import platform

    print(platform.system)
    if platform.system() == "Windows":
        from .app_tools import create_icon
        import pathlib

        logo_fn2 = pathlib.Path(__file__).parent / pathlib.Path("scaffan_icon512.ico")
        create_icon(
            "Scaffan", logo_fn2, conda_env_name="scaffan", package_name="scaffan"
        )

        # logo_fn = op.join(op.dirname(__file__), "scaffan_icon512.ico")
        # import win32com.client
        # shell = win32com.client.Dispatch("WScript.Shell")
        #
        # pth = Path.home()
        # pth = pth / "Desktop" / Path("Scaffan.lnk")
        # shortcut = shell.CreateShortcut(str(pth))
        # # cmd
        # # ln =  "call activate scaffan; {} -m scaffan".format(sys.executable)
        # shortcut.TargetPath = sys.executable
        # shortcut.Arguments = "-m scaffan"
        # # shortcut.TargetPath = cmd
        # # shortcut.Arguments = '/C "call activate scaffan & python -m scaffan" '
        # shortcut.IconLocation = "{},0".format(logo_fn)
        # shortcut.Save()
    pass


@run.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--input-path",
    "-i",
    type=click.Path(exists=True),
    help="Path to input directory with video files.",
    default=None,
)
@click.option(
    "--color", "-c", type=str, help="Annotation color in hexa (#0000FF)", default=None,
)
@click.option(
    "--output-path",
    "-o",
    type=click.Path(),
    help="Path to output directory with video files.",
    default=None,
)
# @click.option(
#     "--log-level",
#     "-ll",
#     # type=,
#     help="Set logging level",
#     default=None,
# )
@click.option(
    "--params",
    "-p",
    multiple=True,
    default="",
    nargs=2,
    help='Set parameter. First argument is path to parameter separated by ";". Second is the value.'
    "python -m scaffan gui -p Processing,Show True",
)
@click.option(
    "--seeds_mm",
    "-smm",
    multiple=True,
    default=None,
    nargs=2,
    help='Set parameter. First argument is path to parameter separated by ";". Second is the value.'
         "python -m scaffan gui -p Processing,Show True",
)
def nogui(input_path, color, output_path, params, seeds_mm):
    # if log_level is not None:
    #     i = logger.add(level=log_level)
    logger.debug(
        f"input path={input_path} color={color}, output_path={output_path}, params={params}"
    )
    mainapp = algorithm.Scaffan()
    logger.debug(f"Scaffan created")
    app_tools.set_parameters_by_path(mainapp.parameters, params)
    # for param in params:
    #     logger.debug(f"param={param}")
    #     mainapp.parameters.param(*param[0].split(";")).setValue(ast.literal_eval(param[1]))

    logger.debug("before input file")
    if input_path is not None:
        mainapp.set_input_file(input_path)
    if output_path is not None:
        mainapp.set_output_dir(output_path)
    if color is not None:
        logger.debug(f"color={color}")
        mainapp.set_annotation_color_selection(color)
    # do float
    logger.debug(f"seeds_mm={seeds_mm}")
    if seeds_mm:
        seeds_mm = [[float(c[0]), float(c[1])] for c in seeds_mm]
        logger.debug(f"seeds_mm readed {seeds_mm}")
    mainapp.run_lobuluses(seeds_mm=seeds_mm)


# def install():
def create_icon(
    app_name: str, icon_filename=None, conda_env_name=None, package_name=None
):
    """

    :param app_name: Used for desktop icon name
    :param icon_filename: absolute path to icon usually:
        pathlib.Path(__file__).parent / pathlib.Path("app_icon512.ico")
    :param conda_env_name: conda environment. The app_name is used if conda_env_name is set to None.
    :param package_name: in `conda -m 'package_name'` . The app_name is used if conda_env_name is set to None.
    :return:
    """
    import platform

    # print(platform.system)
    if conda_env_name is None:
        conda_env_name = app_name

    # if icon_filename is None:
    #     icon_filename = app_name

    if Path(icon_filename).suffix == "":
        icon_filename += ".ico"

    if package_name is None:
        package_name = app_name

    if platform.system() == "Windows":

        # logo_fn2 = pathlib.Path(__file__).parent / pathlib.Path("scaffan_icon512.ico")

        # logo_fn = op.join(op.dirname(__file__), icon_filename)
        logo_fn = icon_filename
        import win32com.client

        shell = win32com.client.Dispatch("WScript.Shell")

        pth = Path.home()
        pth = pth / "Desktop" / Path(f"{app_name}.lnk")
        shortcut = shell.CreateShortcut(str(pth))
        # cmd
        # ln =  "call activate scaffan; {} -m scaffan".format(sys.executable)
        # C:\Windows\System32\cmd.exe /C "call activate anwaapp & pause &  python -m anwa & pause"
        # shortcut.TargetPath = sys.executable
        # shortcut.Arguments = f"-m {app_name}"
        shortcut.TargetPath = "cmd"
        # C:\Windows\System32\cmd.exe /C "call activate anwaapp & pause &  python -m anwa & pause"
        shortcut.Arguments = (
            f'/C "call activate {conda_env_name} & python -m {package_name} & pause" '
        )
        shortcut.IconLocation = "{},0".format(logo_fn)
        shortcut.Save()



import numpy as np


def sigmoidal(x):
    z = 1 / (1 + np.exp(-x))
    return z


def make_norm(imdata, mn, st):
    imdata_norm = (imdata - mn) / st
    return imdata_norm


def hist_mean_std_normalization(imdata):
    pixeldata = imdata.flatten()
    #     plt.hist(pixeldata, bins=40)
    mn = np.mean(pixeldata)
    st = np.std(pixeldata)
    imdata_norm = make_norm(imdata, mn, st)
    return imdata_norm, mn, st


def rescale_intensity_no_limits(img, in_range, out_range=(-0.9, 0.9)):
    x0, x1 = in_range
    y0, y1 = out_range
    k = (y0 - y1) / (x0 - x1)
    q = y0 - k * x0

    img_out = k * img + q
    return img_out


# def rescale_intensity_by_percentile(img, percentile_range=(5, 95), sig_range=(-0.9, 0.9), sig_slope=1):
#     p2, p98 = np.percentile(img, percentile_range)
#     imout = rescale_intensity_no_limits(img, in_range=(p2, p98), out_range=sig_range)
#     return sigmoidal(img * sig_slope)


class RescaleIntensityPercentile:
    def __init__(self):
        self.percentile_range = None
        self.percentile_map_range = None
        self.sig_slope = None
        # self.input_dtype = None
        self.set_parameters()

    def set_parameters(
        self, percentile_range=(5, 95), percentile_map_range=(-0.9, 0.9), sig_slope=1
    ):
        self.percentile_range = percentile_range
        self.percentile_map_range = percentile_map_range
        self.sig_slope = sig_slope

    def calculate_intensity_dependent_parameters(self, img):
        self.percentile_range_values = np.percentile(img, self.percentile_range)
        # self.input_dtype = img.dtype

    def rescale_intensity(self, img):
        input_dtype = img.dtype
        imgout = rescale_intensity_no_limits(
            img,
            in_range=self.percentile_range_values,
            out_range=self.percentile_map_range,
        )
        imgout = sigmoidal(imgout * self.sig_slope)
        if input_dtype == np.uint8:
            imgout = imgout * 255
        return imgout.astype(input_dtype)
